#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class AffectedVerb(object):
    SELECTED = 0
    INSERTED = 1
    UPDATED = 2
    DELETED = 3
    NONE = 4

    _VALUES_TO_NAMES = {
        0: "SELECTED",
        1: "INSERTED",
        2: "UPDATED",
        3: "DELETED",
        4: "NONE",
    }

    _NAMES_TO_VALUES = {
        "SELECTED": 0,
        "INSERTED": 1,
        "UPDATED": 2,
        "DELETED": 3,
        "NONE": 4,
    }


class StatementResponse(object):
    """
    Attributes:
     - statementId
     - statementString
     - numMatchedRows
     - numAffectedRows
     - affectedVerb
     - statementName
    """


    def __init__(self, statementId=None, statementString=None, numMatchedRows=None, numAffectedRows=None, affectedVerb=None, statementName=None,):
        self.statementId = statementId
        self.statementString = statementString
        self.numMatchedRows = numMatchedRows
        self.numAffectedRows = numAffectedRows
        self.affectedVerb = affectedVerb
        self.statementName = statementName

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.statementId = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.statementString = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I64:
                    self.numMatchedRows = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.I64:
                    self.numAffectedRows = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.I32:
                    self.affectedVerb = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.STRING:
                    self.statementName = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('StatementResponse')
        if self.statementId is not None:
            oprot.writeFieldBegin('statementId', TType.STRING, 1)
            oprot.writeString(self.statementId.encode('utf-8') if sys.version_info[0] == 2 else self.statementId)
            oprot.writeFieldEnd()
        if self.statementString is not None:
            oprot.writeFieldBegin('statementString', TType.STRING, 2)
            oprot.writeString(self.statementString.encode('utf-8') if sys.version_info[0] == 2 else self.statementString)
            oprot.writeFieldEnd()
        if self.numMatchedRows is not None:
            oprot.writeFieldBegin('numMatchedRows', TType.I64, 3)
            oprot.writeI64(self.numMatchedRows)
            oprot.writeFieldEnd()
        if self.numAffectedRows is not None:
            oprot.writeFieldBegin('numAffectedRows', TType.I64, 4)
            oprot.writeI64(self.numAffectedRows)
            oprot.writeFieldEnd()
        if self.affectedVerb is not None:
            oprot.writeFieldBegin('affectedVerb', TType.I32, 5)
            oprot.writeI32(self.affectedVerb)
            oprot.writeFieldEnd()
        if self.statementName is not None:
            oprot.writeFieldBegin('statementName', TType.STRING, 6)
            oprot.writeString(self.statementName.encode('utf-8') if sys.version_info[0] == 2 else self.statementName)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.statementId is None:
            raise TProtocolException(message='Required field statementId is unset!')
        if self.numMatchedRows is None:
            raise TProtocolException(message='Required field numMatchedRows is unset!')
        if self.numAffectedRows is None:
            raise TProtocolException(message='Required field numAffectedRows is unset!')
        if self.affectedVerb is None:
            raise TProtocolException(message='Required field affectedVerb is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(StatementResponse)
StatementResponse.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'statementId', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'statementString', 'UTF8', None, ),  # 2
    (3, TType.I64, 'numMatchedRows', None, None, ),  # 3
    (4, TType.I64, 'numAffectedRows', None, None, ),  # 4
    (5, TType.I32, 'affectedVerb', None, None, ),  # 5
    (6, TType.STRING, 'statementName', 'UTF8', None, ),  # 6
)
fix_spec(all_structs)
del all_structs



# -*- coding: utf-8 -*-
# flake8: noqa: E501
import logging
import asyncio
import datetime
import io
import socket
from typing import Optional
import pandas as pd

from pyrdpdb import aiordp
from pyrdpdb.messages.v4.QueryExecuteRequest.ttypes import QueryExecuteRequest

from . import converters, err
from ._compat import IRONPYTHON, PY2, str_type
from .charset import charset_by_name
from .cursors import Cursor
from .utils import _ConnectionContextManager, _ContextManager
from .wireline_protocol import WirelineProtocol
from .utils import timeit

log = logging.getLogger("aiordp.connections")

MAX_PACKET_LEN = 2**24 - 1

if PY2 and not IRONPYTHON:
    # read method of file-like returned by sock.makefile() is very slow.
    # So we copy io-based one from Python 3.
    from ._socketio import SocketIO

    def _makefile(sock, mode):
        return io.BufferedReader(SocketIO(sock, mode))

else:
    # socket.makefile in Python 3 is nice.
    def _makefile(sock, mode):
        return sock.makefile(mode)


DEFAULT_CHARSET = "utf8"


def connect(
    host="localhost",
    port=4333,
    cursorclass=Cursor,
    db=None,
    database=None,
    tprotocol=None,
    write_timeout=None,
    read_timeout=None,
    protocol=None,
    user=None,
    password=None,
    # federation=None,
    catalog=None,
    schema=None,
    charset="",
    encoding=None,
    conv=None,
    connect_timeout=10,
    auth_type="PLAIN",
    echo=False,
    loop=None,
):
    """See connections.Connection.__init__() for information about
    defaults."""
    coro = _connect(
        host=host,
        port=port,
        cursorclass=cursorclass,
        user=user,
        password=password,
        db=db,
        database=database,
        tprotocol=tprotocol,
        write_timeout=write_timeout,
        read_timeout=read_timeout,
        protocol=protocol,
        # federation=federation,
        catalog=catalog,
        schema=schema,
        charset=charset,
        encoding=encoding,
        conv=conv,
        connect_timeout=connect_timeout,
        auth_type=auth_type,
        echo=echo,
        loop=loop,
    )
    return _ConnectionContextManager(coro)


async def _connect(*args, **kwargs):
    conn = Connection(*args, **kwargs)
    await conn._connect()

    # if kwargs.get("catalog"):
    #     await conn._set_catalog(kwargs.get("catalog"))
    # if kwargs.get("schema"):
    #     await conn._set_schema(kwargs.get("schema"))
    return conn


class Connection(object):
    """
    Representation of a socket with a rapidsdb server.
    The proper way to get an instance of this class is to call
    connect().
    Establish a connection to the rapidsdb database. Accepts several
    arguments:
    :param host: Host where the database server is located
    :param user: Username to log in as
    :param password: Password to use.
    :param database: Database to use, None to not use a particular one.
    :param port: rapidsdb port to use, default is usually OK. (default: 4333)
    :param read_timeout: The timeout for reading from the connection in seconds
        (default: None - no timeout)
    :param write_timeout: The timeout for writing to the connection in seconds
        (default: None - no timeout)
    :param charset: Charset you want to use.
    :param conv:
        Conversion dictionary to use instead of the default one.
        This is used to provide custom marshalling and unmarshaling of types.
        See converters.
    :param cursorclass: Custom cursor class to use.
    :param connect_timeout: Timeout before throwing an exception when
        connecting. (default: 10, min: 1, max: 31536000)
    See `Connection
        <https://www.python.org/dev/peps/pep-0249/#connection-objects>`_ in the
        specification.
    """

    _sock = None
    closed = False
    _secure = False

    Warning = err.Warning
    Error = err.Error
    InterfaceError = err.InterfaceError
    DatabaseError = err.DatabaseError
    DataError = err.DataError
    OperationalError = err.OperationalError
    IntegrityError = err.IntegrityError
    InternalError = err.InternalError
    ProgrammingError = err.ProgrammingError
    NotSupportedError = err.NotSupportedError

    # AUTH_TYPE PLAIN / KERBORS
    def __init__(
        self,
        host=None,
        port=4333,
        cursorclass=Cursor,
        db=None,
        database=None,
        tprotocol=None,
        write_timeout=None,
        read_timeout=None,
        protocol=None,
        user=None,
        password=None,
        # federation=None,
        catalog=None,
        schema=None,
        charset="",
        encoding=None,
        conv=None,
        connect_timeout=10,
        auth_type="PLAIN",
        echo=False,
        loop=None,
    ):
        self._loop = loop or asyncio.get_event_loop()
        self._last_usage = self._loop.time()
        self.current_execute_time: float = 0
        self.host = host or "localhost"
        self.port = port or 4333
        self.auth_type = auth_type
        self.user = user
        self.password = password
        self.cursorclass = cursorclass
        self.protocol: Optional[WirelineProtocol] = protocol
        self._echo = echo
        if read_timeout is not None and read_timeout <= 0:
            log.error("read_timeout should be >= 0")
            raise ValueError("read_timeout should be >= 0")
        self._read_timeout = read_timeout
        if write_timeout is not None and write_timeout <= 0:
            log.error("write_timeout should be >= 0")
            raise ValueError("write_timeout should be >= 0")
        self._write_timeout = write_timeout

        connect_timeout = int(connect_timeout)
        if not (0 < connect_timeout <= 31536000):
            raise ValueError("connect_timeout should be >0 and <=31536000")
        self.connect_timeout = connect_timeout or None

        if charset:
            self.charset = charset
        else:
            self.charset = DEFAULT_CHARSET

        self.encoding = charset_by_name(self.charset).encoding

        if conv is None:
            conv = converters.conversions
        self._result: Optional[ResultSet] = None

        self.encoders = {k: v for (k, v) in conv.items() if type(k) is not int}
        self.decoders = {k: v for (k, v) in conv.items() if type(k) is int}

    @property
    def loop(self):
        return self._loop

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.close()
        else:
            self.ensure_closed()
        return

    async def _connect(self, sock=None):
        self.closed = False
        if sock is None:
            sock: socket.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(self.connect_timeout)
            sock.connect((self.host, self.port))
        self._sock = sock
        self._rfile = _makefile(sock, "rb")
        if self.protocol is None:
            self.protocol = WirelineProtocol(self)
        self.connected_time = self._loop.time()
        await self._login()

    async def _login(self):
        await self.protocol.login()

    async def _set_catalog(self, catalog_name):
        maxlength = 128

        if catalog_name and (len(catalog_name) > maxlength or len(catalog_name) < 1):
            # TODO: raise execption
            pass

        async with self.cursor() as cursor:
            stmt = f"SET CATALOG {catalog_name};"
            await cursor.execute(stmt, quiet=True)

    async def _set_schema(self, schema_name):
        maxlength = 128

        if schema_name and (len(schema_name) > maxlength or len(schema_name) < 1):
            # TODO: raise execption
            pass

        async with self.cursor() as cursor:
            sql = f"SET SCHEMA {schema_name};"
            await cursor.execute(sql, quiet=True)

    def ensure_closed(self):
        """Send quit command and then close socket connection"""
        if self._sock is None:
            # connection has been closed
            return
        self.close()

    def escape(self, obj, mapping=None):
        """
        Escape whatever value you pass to it.
        Non-standard, for internal use; do not use this in your applications.
        """
        if isinstance(obj, str_type):
            return "'" + converters.escape_string(obj) + "'"
        if isinstance(obj, (bytes, bytearray)):
            return converters.escape_bytes(obj)
        return converters.escape_item(obj, self.charset, mapping=mapping)

    def literal(self, obj):
        """
        Alias for escape()
        Non-standard, for internal use; do not use this in your applications.
        """
        return self.escape(obj, self.encoders)

    def __del__(self):
        if self._sock:
            self.close()
        if self._sock:
            log.warning(f"Unclosed connection {self!r}", ResourceWarning)

    def close(self):
        """
        Close connection .
        """
        if self.closed:
            raise err.Error("Already closed")
        self.closed = True
        if self._sock is None:
            return
        self._force_close()

    def _force_close(self):
        if self._sock:
            try:
                self._sock.close()
            except Exception:
                pass
        self._sock = None
        self._rfile = None

    async def commit(self):
        """
        Does nothing, required by DB API.
        """
        pass

    async def rollback(self):
        """
        Does nothing, required by DB API.
        """
        pass

    async def begin(self):
        """
        Does nothing, required by DB API.
        """
        pass

    @timeit
    async def to_pandas(self, sql) -> pd.DataFrame:
        dfs = []
        await self.protocol.query(sql)

        while await self.protocol.to_pandas():
            dfs.append(self.protocol._df)
        if self.protocol._df:
            dfs.append(self.protocol._df)
        df = pd.concat(dfs, ignore_index=True)

        return df

    @timeit
    async def to_pandas2(self, sql) -> pd.DataFrame:
        dfs = []
        await self.protocol.query(sql)

        result = ResultSet(self.protocol)
        row_generator = await result.read_generator()
        columns = [desc[0] for desc in self.protocol.description]
        for block in row_generator:
            df = pd.DataFrame(block, columns=columns)
            dfs.append(df)

        df = pd.concat(dfs, ignore_index=True)

        return df

    # @timeit
    async def query(self, sql, unbuffered=False, generator=False):
        start = datetime.datetime.now()
        await self.protocol.query(sql)
        end = datetime.datetime.now()
        self.current_execute_time = round((end - start).total_seconds(), 2)

        if not generator:
            self._affected_rows = await self._read_query_result(unbuffered)
        else:
            self._affected_rows = self._read_query_result_generator(unbuffered)

    async def next_result(self, unbuffered=False):
        self._affected_rows = await self._read_query_result(unbuffered=unbuffered)
        return self._affected_rows

    def cursor(self, *cursors):
        """Instantiates and returns a cursor

        By default, :class:`Cursor` is returned. It is possible to also give a
        custom cursor through the cursor_class parameter, but it needs to
        be a subclass  of :class:`Cursor`.

        Note: this function should be treated as async function. Otherwise,
        it would just return a '_ContextManager' object.

        :param cursor: custom cursor class.
        :returns: instance of cursor, by default :class:`Cursor`
        :raises TypeError: cursor_class is not a subclass of Cursor.
        """
        # self._ensure_alive()
        self._last_usage = self._loop.time()
        try:
            if cursors and any(not issubclass(cursor, Cursor) for cursor in cursors):
                raise TypeError("Custom cursor must be subclass of Cursor")
        except TypeError:
            raise TypeError("Custom cursor must be subclass of Cursor")
        if cursors and len(cursors) == 1:
            cur = cursors[0](self, self._echo)
        elif cursors:
            cursor_name = (
                "".join(map(lambda x: x.__name__, cursors)).replace("Cursor", "")
                + "Cursor"
            )
            cursor_class = type(cursor_name, cursors, {})
            cur = cursor_class(self, self._echo)
        else:
            cur = self.cursorclass(self, self._echo)
        fut = self._loop.create_future()
        fut.set_result(cur)
        return _ContextManager(fut)

    async def _read_query_result(self, unbuffered=False):
        self._result = None
        if unbuffered:
            try:
                result = ResultSet(self.protocol)
                result.init_unbuffered_query()
            except:
                result.unbuffered_active = False
                result.protocol = None
                raise
        else:
            result = ResultSet(self.protocol)
            await result.read()

        self._result = result
        return result.affected_rows

    async def _read_query_result_generator(self, unbuffered=False) -> Optional[int]:
        self._result = None

        try:
            result = ResultSet(self.protocol)

            if unbuffered:
                result.init_unbuffered_query()
            else:
                row_generator = result.read_generator()

                for block in row_generator:
                    # print("generating block")
                    current_chunk = []
                    for row in block:
                        current_chunk.append(row)
                    result.rows = current_chunk

                    self._result = result
            return result.affected_rows
        except Exception as e:
            if unbuffered:
                result.unbuffered_active = False
                result.protocol = None
            raise e


class ResultSet(object):
    def __init__(self, protocol: WirelineProtocol):
        self.protocol: Optional[WirelineProtocol] = protocol
        self.affected_rows = None
        self.rows = None
        self.unbuffered = False
        self.description = None
        self.field_count = 0
        self.err = None
        self.has_next = None  # should be bool, decide if there is next row

        if isinstance(self.protocol, WirelineProtocol):
            self.has_next = self.protocol.has_next

    # buffered result, cache all the data to the client
    async def read(self):
        rows = []

        try:
            while await self.protocol.get_next_set():
                rows.extend(self.protocol.chunk)

            if self.protocol.chunk:
                rows.extend(self.protocol.chunk)

            self.has_next = self.protocol.has_next
        except Exception:
            log.debug("No more data")
            self.has_next = False

        self.affected_rows = len(rows)
        self.rows = rows
        self.description = self.protocol.description
        self.field_count = self.protocol.numColumns
        self.err = self.protocol.errstr
        # self.protocol = None

    async def read_generator(self):

        async def merge_chunk():
            nonlocal count
            _chunk = self.protocol.chunk
            if _chunk:
                count += len(_chunk)
                self.affected_rows = count
                log.debug(
                    "affected_rows: {}, count: {}".format(self.affected_rows, count)
                )
                yield _chunk

        try:
            self.description = self.protocol.description
            self.field_count = self.protocol.numColumns
            self.err = self.protocol.errstr

            count = 0

            async for _chunk in merge_chunk():
                yield _chunk

            while await self.protocol.get_next_set():
                async for _chunk in merge_chunk():
                    yield _chunk

            async for _chunk in merge_chunk():
                yield _chunk

            self.has_next = self.protocol.has_next

        except Exception as e:
            log.debug("No more data: %s", str(e))
            self.has_next = False

    def init_unbuffered_query(self):
        self.description = self.protocol.description
        self.field_count = self.protocol.numColumns
        self.err = self.protocol.errstr
        self.affected_rows = -1
        self.unbuffered = True
        if self.err:
            raise aiordp.ProgrammingError(self.err)

    async def _finish_unbuffered_query(self):
        while self.has_next:
            await self.read()
        self.protocol = None

    async def read_next_row(self):
        if not self.unbuffered or self.protocol is None:
            return

        has_next = await self.protocol.get_next_row()
        if self.protocol and has_next:
            row = self.protocol.row
        else:
            self.protocol = None
            row = None

        self.rows = (row,)

        if row is not None:
            self.affected_rows = 1

        return row



#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class DroppedMessageInfo(object):
    """
    Attributes:
     - messageSize
     - messageType
    """


    def __init__(self, messageSize=None, messageType=None,):
        self.messageSize = messageSize
        self.messageType = messageType

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.messageSize = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.messageType = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('DroppedMessageInfo')
        if self.messageSize is not None:
            oprot.writeFieldBegin('messageSize', TType.I32, 1)
            oprot.writeI32(self.messageSize)
            oprot.writeFieldEnd()
        if self.messageType is not None:
            oprot.writeFieldBegin('messageType', TType.I32, 2)
            oprot.writeI32(self.messageType)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.messageSize is None:
            raise TProtocolException(message='Required field messageSize is unset!')
        if self.messageType is None:
            raise TProtocolException(message='Required field messageType is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class SessionDroppedMessagesResponse(object):
    """
    Attributes:
     - droppedMessages
     - missingDroppedMessages
    """


    def __init__(self, droppedMessages=None, missingDroppedMessages=None,):
        self.droppedMessages = droppedMessages
        self.missingDroppedMessages = missingDroppedMessages

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.droppedMessages = []
                    (_etype3, _size0) = iprot.readListBegin()
                    for _i4 in range(_size0):
                        _elem5 = DroppedMessageInfo()
                        _elem5.read(iprot)
                        self.droppedMessages.append(_elem5)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.BOOL:
                    self.missingDroppedMessages = iprot.readBool()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('SessionDroppedMessagesResponse')
        if self.droppedMessages is not None:
            oprot.writeFieldBegin('droppedMessages', TType.LIST, 1)
            oprot.writeListBegin(TType.STRUCT, len(self.droppedMessages))
            for iter6 in self.droppedMessages:
                iter6.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.missingDroppedMessages is not None:
            oprot.writeFieldBegin('missingDroppedMessages', TType.BOOL, 2)
            oprot.writeBool(self.missingDroppedMessages)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.missingDroppedMessages is None:
            raise TProtocolException(message='Required field missingDroppedMessages is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(DroppedMessageInfo)
DroppedMessageInfo.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'messageSize', None, None, ),  # 1
    (2, TType.I32, 'messageType', None, None, ),  # 2
)
all_structs.append(SessionDroppedMessagesResponse)
SessionDroppedMessagesResponse.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'droppedMessages', (TType.STRUCT, [DroppedMessageInfo, None], False), None, ),  # 1
    (2, TType.BOOL, 'missingDroppedMessages', None, None, ),  # 2
)
fix_spec(all_structs)
del all_structs



#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class IntervalType(object):
    YEAR_MONTH = 0
    DAY_TIME = 1

    _VALUES_TO_NAMES = {
        0: "YEAR_MONTH",
        1: "DAY_TIME",
    }

    _NAMES_TO_VALUES = {
        "YEAR_MONTH": 0,
        "DAY_TIME": 1,
    }


class Timestamp(object):
    """
    Attributes:
     - year
     - month
     - day
     - hour
     - minute
     - second
     - nanoseconds
     - offsetHours
     - offsetMinutes
     - tzName
    """


    def __init__(self, year=None, month=None, day=None, hour=None, minute=None, second=None, nanoseconds=None, offsetHours=None, offsetMinutes=None, tzName=None,):
        self.year = year
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second
        self.nanoseconds = nanoseconds
        self.offsetHours = offsetHours
        self.offsetMinutes = offsetMinutes
        self.tzName = tzName

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.year = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I16:
                    self.month = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I16:
                    self.day = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.I16:
                    self.hour = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.I16:
                    self.minute = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.I16:
                    self.second = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 7:
                if ftype == TType.I32:
                    self.nanoseconds = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 8:
                if ftype == TType.BYTE:
                    self.offsetHours = iprot.readByte()
                else:
                    iprot.skip(ftype)
            elif fid == 9:
                if ftype == TType.BYTE:
                    self.offsetMinutes = iprot.readByte()
                else:
                    iprot.skip(ftype)
            elif fid == 10:
                if ftype == TType.STRING:
                    self.tzName = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Timestamp')
        if self.year is not None:
            oprot.writeFieldBegin('year', TType.I32, 1)
            oprot.writeI32(self.year)
            oprot.writeFieldEnd()
        if self.month is not None:
            oprot.writeFieldBegin('month', TType.I16, 2)
            oprot.writeI16(self.month)
            oprot.writeFieldEnd()
        if self.day is not None:
            oprot.writeFieldBegin('day', TType.I16, 3)
            oprot.writeI16(self.day)
            oprot.writeFieldEnd()
        if self.hour is not None:
            oprot.writeFieldBegin('hour', TType.I16, 4)
            oprot.writeI16(self.hour)
            oprot.writeFieldEnd()
        if self.minute is not None:
            oprot.writeFieldBegin('minute', TType.I16, 5)
            oprot.writeI16(self.minute)
            oprot.writeFieldEnd()
        if self.second is not None:
            oprot.writeFieldBegin('second', TType.I16, 6)
            oprot.writeI16(self.second)
            oprot.writeFieldEnd()
        if self.nanoseconds is not None:
            oprot.writeFieldBegin('nanoseconds', TType.I32, 7)
            oprot.writeI32(self.nanoseconds)
            oprot.writeFieldEnd()
        if self.offsetHours is not None:
            oprot.writeFieldBegin('offsetHours', TType.BYTE, 8)
            oprot.writeByte(self.offsetHours)
            oprot.writeFieldEnd()
        if self.offsetMinutes is not None:
            oprot.writeFieldBegin('offsetMinutes', TType.BYTE, 9)
            oprot.writeByte(self.offsetMinutes)
            oprot.writeFieldEnd()
        if self.tzName is not None:
            oprot.writeFieldBegin('tzName', TType.STRING, 10)
            oprot.writeString(self.tzName.encode('utf-8') if sys.version_info[0] == 2 else self.tzName)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.year is None:
            raise TProtocolException(message='Required field year is unset!')
        if self.month is None:
            raise TProtocolException(message='Required field month is unset!')
        if self.day is None:
            raise TProtocolException(message='Required field day is unset!')
        if self.hour is None:
            raise TProtocolException(message='Required field hour is unset!')
        if self.minute is None:
            raise TProtocolException(message='Required field minute is unset!')
        if self.second is None:
            raise TProtocolException(message='Required field second is unset!')
        if self.nanoseconds is None:
            raise TProtocolException(message='Required field nanoseconds is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Date(object):
    """
    Attributes:
     - year
     - month
     - day
     - offsetHours
     - offsetMinutes
     - tzName
    """


    def __init__(self, year=None, month=None, day=None, offsetHours=None, offsetMinutes=None, tzName=None,):
        self.year = year
        self.month = month
        self.day = day
        self.offsetHours = offsetHours
        self.offsetMinutes = offsetMinutes
        self.tzName = tzName

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.year = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I16:
                    self.month = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I16:
                    self.day = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.BYTE:
                    self.offsetHours = iprot.readByte()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.BYTE:
                    self.offsetMinutes = iprot.readByte()
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.STRING:
                    self.tzName = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Date')
        if self.year is not None:
            oprot.writeFieldBegin('year', TType.I32, 1)
            oprot.writeI32(self.year)
            oprot.writeFieldEnd()
        if self.month is not None:
            oprot.writeFieldBegin('month', TType.I16, 2)
            oprot.writeI16(self.month)
            oprot.writeFieldEnd()
        if self.day is not None:
            oprot.writeFieldBegin('day', TType.I16, 3)
            oprot.writeI16(self.day)
            oprot.writeFieldEnd()
        if self.offsetHours is not None:
            oprot.writeFieldBegin('offsetHours', TType.BYTE, 4)
            oprot.writeByte(self.offsetHours)
            oprot.writeFieldEnd()
        if self.offsetMinutes is not None:
            oprot.writeFieldBegin('offsetMinutes', TType.BYTE, 5)
            oprot.writeByte(self.offsetMinutes)
            oprot.writeFieldEnd()
        if self.tzName is not None:
            oprot.writeFieldBegin('tzName', TType.STRING, 6)
            oprot.writeString(self.tzName.encode('utf-8') if sys.version_info[0] == 2 else self.tzName)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.year is None:
            raise TProtocolException(message='Required field year is unset!')
        if self.month is None:
            raise TProtocolException(message='Required field month is unset!')
        if self.day is None:
            raise TProtocolException(message='Required field day is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Time(object):
    """
    Attributes:
     - hour
     - minute
     - second
     - nanoseconds
     - offsetHours
     - offsetMinutes
     - tzName
    """


    def __init__(self, hour=None, minute=None, second=None, nanoseconds=None, offsetHours=None, offsetMinutes=None, tzName=None,):
        self.hour = hour
        self.minute = minute
        self.second = second
        self.nanoseconds = nanoseconds
        self.offsetHours = offsetHours
        self.offsetMinutes = offsetMinutes
        self.tzName = tzName

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I16:
                    self.hour = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I16:
                    self.minute = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I16:
                    self.second = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.I32:
                    self.nanoseconds = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.BYTE:
                    self.offsetHours = iprot.readByte()
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.BYTE:
                    self.offsetMinutes = iprot.readByte()
                else:
                    iprot.skip(ftype)
            elif fid == 7:
                if ftype == TType.STRING:
                    self.tzName = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Time')
        if self.hour is not None:
            oprot.writeFieldBegin('hour', TType.I16, 1)
            oprot.writeI16(self.hour)
            oprot.writeFieldEnd()
        if self.minute is not None:
            oprot.writeFieldBegin('minute', TType.I16, 2)
            oprot.writeI16(self.minute)
            oprot.writeFieldEnd()
        if self.second is not None:
            oprot.writeFieldBegin('second', TType.I16, 3)
            oprot.writeI16(self.second)
            oprot.writeFieldEnd()
        if self.nanoseconds is not None:
            oprot.writeFieldBegin('nanoseconds', TType.I32, 4)
            oprot.writeI32(self.nanoseconds)
            oprot.writeFieldEnd()
        if self.offsetHours is not None:
            oprot.writeFieldBegin('offsetHours', TType.BYTE, 5)
            oprot.writeByte(self.offsetHours)
            oprot.writeFieldEnd()
        if self.offsetMinutes is not None:
            oprot.writeFieldBegin('offsetMinutes', TType.BYTE, 6)
            oprot.writeByte(self.offsetMinutes)
            oprot.writeFieldEnd()
        if self.tzName is not None:
            oprot.writeFieldBegin('tzName', TType.STRING, 7)
            oprot.writeString(self.tzName.encode('utf-8') if sys.version_info[0] == 2 else self.tzName)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.hour is None:
            raise TProtocolException(message='Required field hour is unset!')
        if self.minute is None:
            raise TProtocolException(message='Required field minute is unset!')
        if self.second is None:
            raise TProtocolException(message='Required field second is unset!')
        if self.nanoseconds is None:
            raise TProtocolException(message='Required field nanoseconds is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Interval(object):
    """
    Attributes:
     - type
     - year
     - month
     - day
     - hour
     - minute
     - second
     - nanoseconds
    """


    def __init__(self, type=None, year=None, month=None, day=None, hour=None, minute=None, second=None, nanoseconds=None,):
        self.type = type
        self.year = year
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second
        self.nanoseconds = nanoseconds

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.type = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I64:
                    self.year = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I64:
                    self.month = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.I64:
                    self.day = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.I64:
                    self.hour = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.I64:
                    self.minute = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 7:
                if ftype == TType.I64:
                    self.second = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 8:
                if ftype == TType.I64:
                    self.nanoseconds = iprot.readI64()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Interval')
        if self.type is not None:
            oprot.writeFieldBegin('type', TType.I32, 1)
            oprot.writeI32(self.type)
            oprot.writeFieldEnd()
        if self.year is not None:
            oprot.writeFieldBegin('year', TType.I64, 2)
            oprot.writeI64(self.year)
            oprot.writeFieldEnd()
        if self.month is not None:
            oprot.writeFieldBegin('month', TType.I64, 3)
            oprot.writeI64(self.month)
            oprot.writeFieldEnd()
        if self.day is not None:
            oprot.writeFieldBegin('day', TType.I64, 4)
            oprot.writeI64(self.day)
            oprot.writeFieldEnd()
        if self.hour is not None:
            oprot.writeFieldBegin('hour', TType.I64, 5)
            oprot.writeI64(self.hour)
            oprot.writeFieldEnd()
        if self.minute is not None:
            oprot.writeFieldBegin('minute', TType.I64, 6)
            oprot.writeI64(self.minute)
            oprot.writeFieldEnd()
        if self.second is not None:
            oprot.writeFieldBegin('second', TType.I64, 7)
            oprot.writeI64(self.second)
            oprot.writeFieldEnd()
        if self.nanoseconds is not None:
            oprot.writeFieldBegin('nanoseconds', TType.I64, 8)
            oprot.writeI64(self.nanoseconds)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.type is None:
            raise TProtocolException(message='Required field type is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class ColumnValue(object):
    """
    Attributes:
     - isNull
     - booleanVal
     - tinyIntVal
     - smallIntVal
     - integerVal
     - bigintVal
     - doubleVal
     - bigDecimalVal
     - timestampVal
     - timeVal
     - dateVal
     - intervalVal
     - stringVal
     - binaryVal
    """


    def __init__(self, isNull=None, booleanVal=None, tinyIntVal=None, smallIntVal=None, integerVal=None, bigintVal=None, doubleVal=None, bigDecimalVal=None, timestampVal=None, timeVal=None, dateVal=None, intervalVal=None, stringVal=None, binaryVal=None,):
        self.isNull = isNull
        self.booleanVal = booleanVal
        self.tinyIntVal = tinyIntVal
        self.smallIntVal = smallIntVal
        self.integerVal = integerVal
        self.bigintVal = bigintVal
        self.doubleVal = doubleVal
        self.bigDecimalVal = bigDecimalVal
        self.timestampVal = timestampVal
        self.timeVal = timeVal
        self.dateVal = dateVal
        self.intervalVal = intervalVal
        self.stringVal = stringVal
        self.binaryVal = binaryVal

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.BOOL:
                    self.isNull = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.BOOL:
                    self.booleanVal = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I16:
                    self.tinyIntVal = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.I16:
                    self.smallIntVal = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.I32:
                    self.integerVal = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.I64:
                    self.bigintVal = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 7:
                if ftype == TType.DOUBLE:
                    self.doubleVal = iprot.readDouble()
                else:
                    iprot.skip(ftype)
            elif fid == 8:
                if ftype == TType.STRING:
                    self.bigDecimalVal = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 9:
                if ftype == TType.STRUCT:
                    self.timestampVal = Timestamp()
                    self.timestampVal.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 10:
                if ftype == TType.STRUCT:
                    self.timeVal = Time()
                    self.timeVal.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 11:
                if ftype == TType.STRUCT:
                    self.dateVal = Date()
                    self.dateVal.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 12:
                if ftype == TType.STRUCT:
                    self.intervalVal = Interval()
                    self.intervalVal.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 13:
                if ftype == TType.STRING:
                    self.stringVal = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 14:
                if ftype == TType.STRING:
                    self.binaryVal = iprot.readBinary()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('ColumnValue')
        if self.isNull is not None:
            oprot.writeFieldBegin('isNull', TType.BOOL, 1)
            oprot.writeBool(self.isNull)
            oprot.writeFieldEnd()
        if self.booleanVal is not None:
            oprot.writeFieldBegin('booleanVal', TType.BOOL, 2)
            oprot.writeBool(self.booleanVal)
            oprot.writeFieldEnd()
        if self.tinyIntVal is not None:
            oprot.writeFieldBegin('tinyIntVal', TType.I16, 3)
            oprot.writeI16(self.tinyIntVal)
            oprot.writeFieldEnd()
        if self.smallIntVal is not None:
            oprot.writeFieldBegin('smallIntVal', TType.I16, 4)
            oprot.writeI16(self.smallIntVal)
            oprot.writeFieldEnd()
        if self.integerVal is not None:
            oprot.writeFieldBegin('integerVal', TType.I32, 5)
            oprot.writeI32(self.integerVal)
            oprot.writeFieldEnd()
        if self.bigintVal is not None:
            oprot.writeFieldBegin('bigintVal', TType.I64, 6)
            oprot.writeI64(self.bigintVal)
            oprot.writeFieldEnd()
        if self.doubleVal is not None:
            oprot.writeFieldBegin('doubleVal', TType.DOUBLE, 7)
            oprot.writeDouble(self.doubleVal)
            oprot.writeFieldEnd()
        if self.bigDecimalVal is not None:
            oprot.writeFieldBegin('bigDecimalVal', TType.STRING, 8)
            oprot.writeString(self.bigDecimalVal.encode('utf-8') if sys.version_info[0] == 2 else self.bigDecimalVal)
            oprot.writeFieldEnd()
        if self.timestampVal is not None:
            oprot.writeFieldBegin('timestampVal', TType.STRUCT, 9)
            self.timestampVal.write(oprot)
            oprot.writeFieldEnd()
        if self.timeVal is not None:
            oprot.writeFieldBegin('timeVal', TType.STRUCT, 10)
            self.timeVal.write(oprot)
            oprot.writeFieldEnd()
        if self.dateVal is not None:
            oprot.writeFieldBegin('dateVal', TType.STRUCT, 11)
            self.dateVal.write(oprot)
            oprot.writeFieldEnd()
        if self.intervalVal is not None:
            oprot.writeFieldBegin('intervalVal', TType.STRUCT, 12)
            self.intervalVal.write(oprot)
            oprot.writeFieldEnd()
        if self.stringVal is not None:
            oprot.writeFieldBegin('stringVal', TType.STRING, 13)
            oprot.writeString(self.stringVal.encode('utf-8') if sys.version_info[0] == 2 else self.stringVal)
            oprot.writeFieldEnd()
        if self.binaryVal is not None:
            oprot.writeFieldBegin('binaryVal', TType.STRING, 14)
            oprot.writeBinary(self.binaryVal)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(Timestamp)
Timestamp.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'year', None, None, ),  # 1
    (2, TType.I16, 'month', None, None, ),  # 2
    (3, TType.I16, 'day', None, None, ),  # 3
    (4, TType.I16, 'hour', None, None, ),  # 4
    (5, TType.I16, 'minute', None, None, ),  # 5
    (6, TType.I16, 'second', None, None, ),  # 6
    (7, TType.I32, 'nanoseconds', None, None, ),  # 7
    (8, TType.BYTE, 'offsetHours', None, None, ),  # 8
    (9, TType.BYTE, 'offsetMinutes', None, None, ),  # 9
    (10, TType.STRING, 'tzName', 'UTF8', None, ),  # 10
)
all_structs.append(Date)
Date.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'year', None, None, ),  # 1
    (2, TType.I16, 'month', None, None, ),  # 2
    (3, TType.I16, 'day', None, None, ),  # 3
    (4, TType.BYTE, 'offsetHours', None, None, ),  # 4
    (5, TType.BYTE, 'offsetMinutes', None, None, ),  # 5
    (6, TType.STRING, 'tzName', 'UTF8', None, ),  # 6
)
all_structs.append(Time)
Time.thrift_spec = (
    None,  # 0
    (1, TType.I16, 'hour', None, None, ),  # 1
    (2, TType.I16, 'minute', None, None, ),  # 2
    (3, TType.I16, 'second', None, None, ),  # 3
    (4, TType.I32, 'nanoseconds', None, None, ),  # 4
    (5, TType.BYTE, 'offsetHours', None, None, ),  # 5
    (6, TType.BYTE, 'offsetMinutes', None, None, ),  # 6
    (7, TType.STRING, 'tzName', 'UTF8', None, ),  # 7
)
all_structs.append(Interval)
Interval.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'type', None, None, ),  # 1
    (2, TType.I64, 'year', None, None, ),  # 2
    (3, TType.I64, 'month', None, None, ),  # 3
    (4, TType.I64, 'day', None, None, ),  # 4
    (5, TType.I64, 'hour', None, None, ),  # 5
    (6, TType.I64, 'minute', None, None, ),  # 6
    (7, TType.I64, 'second', None, None, ),  # 7
    (8, TType.I64, 'nanoseconds', None, None, ),  # 8
)
all_structs.append(ColumnValue)
ColumnValue.thrift_spec = (
    None,  # 0
    (1, TType.BOOL, 'isNull', None, None, ),  # 1
    (2, TType.BOOL, 'booleanVal', None, None, ),  # 2
    (3, TType.I16, 'tinyIntVal', None, None, ),  # 3
    (4, TType.I16, 'smallIntVal', None, None, ),  # 4
    (5, TType.I32, 'integerVal', None, None, ),  # 5
    (6, TType.I64, 'bigintVal', None, None, ),  # 6
    (7, TType.DOUBLE, 'doubleVal', None, None, ),  # 7
    (8, TType.STRING, 'bigDecimalVal', 'UTF8', None, ),  # 8
    (9, TType.STRUCT, 'timestampVal', [Timestamp, None], None, ),  # 9
    (10, TType.STRUCT, 'timeVal', [Time, None], None, ),  # 10
    (11, TType.STRUCT, 'dateVal', [Date, None], None, ),  # 11
    (12, TType.STRUCT, 'intervalVal', [Interval, None], None, ),  # 12
    (13, TType.STRING, 'stringVal', 'UTF8', None, ),  # 13
    (14, TType.STRING, 'binaryVal', 'BINARY', None, ),  # 14
)
fix_spec(all_structs)
del all_structs



#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class DataType(object):
    NULL = 0
    BOOLEAN = 1
    TINYINT = 2
    SMALLINT = 3
    INTEGER = 4
    BIGINT = 5
    DECIMAL = 6
    FLOAT = 7
    DATE = 8
    TIME = 9
    TIMESTAMP = 10
    INTERVAL = 11
    VARCHAR = 12
    VARBINARY = 13

    _VALUES_TO_NAMES = {
        0: "NULL",
        1: "BOOLEAN",
        2: "TINYINT",
        3: "SMALLINT",
        4: "INTEGER",
        5: "BIGINT",
        6: "DECIMAL",
        7: "FLOAT",
        8: "DATE",
        9: "TIME",
        10: "TIMESTAMP",
        11: "INTERVAL",
        12: "VARCHAR",
        13: "VARBINARY",
    }

    _NAMES_TO_VALUES = {
        "NULL": 0,
        "BOOLEAN": 1,
        "TINYINT": 2,
        "SMALLINT": 3,
        "INTEGER": 4,
        "BIGINT": 5,
        "DECIMAL": 6,
        "FLOAT": 7,
        "DATE": 8,
        "TIME": 9,
        "TIMESTAMP": 10,
        "INTERVAL": 11,
        "VARCHAR": 12,
        "VARBINARY": 13,
    }


class ColumnMetadata(object):
    """
    Attributes:
     - columnName
     - columnType
     - precision
     - scale
     - isNullable
     - charsetName
     - collationName
     - isAutoIncrement
    """


    def __init__(self, columnName=None, columnType=None, precision=None, scale=None, isNullable=None, charsetName=None, collationName=None, isAutoIncrement=None,):
        self.columnName = columnName
        self.columnType = columnType
        self.precision = precision
        self.scale = scale
        self.isNullable = isNullable
        self.charsetName = charsetName
        self.collationName = collationName
        self.isAutoIncrement = isAutoIncrement

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.columnName = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.columnType = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I32:
                    self.precision = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.I32:
                    self.scale = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.BOOL:
                    self.isNullable = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.STRING:
                    self.charsetName = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 7:
                if ftype == TType.STRING:
                    self.collationName = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 8:
                if ftype == TType.BOOL:
                    self.isAutoIncrement = iprot.readBool()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('ColumnMetadata')
        if self.columnName is not None:
            oprot.writeFieldBegin('columnName', TType.STRING, 1)
            oprot.writeString(self.columnName.encode('utf-8') if sys.version_info[0] == 2 else self.columnName)
            oprot.writeFieldEnd()
        if self.columnType is not None:
            oprot.writeFieldBegin('columnType', TType.I32, 2)
            oprot.writeI32(self.columnType)
            oprot.writeFieldEnd()
        if self.precision is not None:
            oprot.writeFieldBegin('precision', TType.I32, 3)
            oprot.writeI32(self.precision)
            oprot.writeFieldEnd()
        if self.scale is not None:
            oprot.writeFieldBegin('scale', TType.I32, 4)
            oprot.writeI32(self.scale)
            oprot.writeFieldEnd()
        if self.isNullable is not None:
            oprot.writeFieldBegin('isNullable', TType.BOOL, 5)
            oprot.writeBool(self.isNullable)
            oprot.writeFieldEnd()
        if self.charsetName is not None:
            oprot.writeFieldBegin('charsetName', TType.STRING, 6)
            oprot.writeString(self.charsetName.encode('utf-8') if sys.version_info[0] == 2 else self.charsetName)
            oprot.writeFieldEnd()
        if self.collationName is not None:
            oprot.writeFieldBegin('collationName', TType.STRING, 7)
            oprot.writeString(self.collationName.encode('utf-8') if sys.version_info[0] == 2 else self.collationName)
            oprot.writeFieldEnd()
        if self.isAutoIncrement is not None:
            oprot.writeFieldBegin('isAutoIncrement', TType.BOOL, 8)
            oprot.writeBool(self.isAutoIncrement)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.columnName is None:
            raise TProtocolException(message='Required field columnName is unset!')
        if self.columnType is None:
            raise TProtocolException(message='Required field columnType is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(ColumnMetadata)
ColumnMetadata.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'columnName', 'UTF8', None, ),  # 1
    (2, TType.I32, 'columnType', None, None, ),  # 2
    (3, TType.I32, 'precision', None, None, ),  # 3
    (4, TType.I32, 'scale', None, None, ),  # 4
    (5, TType.BOOL, 'isNullable', None, None, ),  # 5
    (6, TType.STRING, 'charsetName', 'UTF8', None, ),  # 6
    (7, TType.STRING, 'collationName', 'UTF8', None, ),  # 7
    (8, TType.BOOL, 'isAutoIncrement', None, None, ),  # 8
)
fix_spec(all_structs)
del all_structs



# flake8: noqa: E501
from sqlalchemy import types as sqltypes

"""
Type		|    Default interpretation of size / scale / precision
--------------------------------------------------------------------
INTEGER	    |    Precision in decimal digits (if unspecified: 18)
DECIMAL	    |    Precision in decimal digits, scale in decimal digits (if unspecified: 19, 0)
FLOAT		|    Size of mantissa in binary digits (if unspecified: 53)
TIMESTAMP   |
VARCHAR	    |    Physical size in bytes (if unspecifed: variable up to 2G)
VARBINARY	|    Physical size in bytes (if unspecified: variable up to 2G)
Boolean     |
CHAR        |    RDP do not support CHAR, but RDP have a bug can have the deep datasource CHAR type,  I have no idea on this.
"""


class BOOLEAN(sqltypes.BOOLEAN):
    __visit_name__ = "BOOLEAN"


class _NumericType(object):
    """Base for MySQL numeric types.

    This is the base both for NUMERIC as well as INTEGER, hence
    it's a mixin.

    """

    def __init__(self, unsigned=False, zerofill=False, **kw):
        self.unsigned = unsigned
        self.zerofill = zerofill
        super(_NumericType, self).__init__(**kw)


# Do not have TINYINT
class TINYINT(sqltypes.INTEGER):
    __visit_name__ = "TINYINT"


class SMALLINT(sqltypes.SMALLINT):
    __visit_name__ = "SMALLINT"


class INTEGER(sqltypes.INTEGER):
    __visit_name__ = "INTEGER"


class BIGINT(sqltypes.BIGINT):
    __visit_name__ = "INTEGER"


class DECIMAL(sqltypes.DECIMAL):
    __visit_name__ = "DECIMAL"


# class DECIMAL(_NumericType, sqltypes.DECIMAL):
#     """MySQL DECIMAL type."""

#     __visit_name__ = "DECIMAL"

#     def __init__(self, precision=None, scale=None, asdecimal=True, **kw):
#         """Construct a DECIMAL.

#         :param precision: Total digits in this number.  If scale and precision
#           are both None, values are stored to limits allowed by the server.

#         :param scale: The number of digits after the decimal point.

#         :param unsigned: a boolean, optional.

#         :param zerofill: Optional. If true, values will be stored as strings
#           left-padded with zeros. Note that this does not effect the values
#           returned by the underlying database API, which continue to be
#           numeric.

#         """
#         super(DECIMAL, self).__init__(
#             precision=precision, scale=scale, asdecimal=asdecimal, **kw
#         )


class FLOAT(sqltypes.FLOAT):
    __visit_name__ = "FLOAT"


class DATE(sqltypes.DATE):
    __visit_name__ = "DATE"


class TIME(sqltypes.TIME):
    __visit_name__ = "TIME"


class TIMESTAMP(sqltypes.TIMESTAMP):
    __visit_name__ = "TIMESTAMP"


class DATETIME(sqltypes.DateTime):
    __visit_name__ = "DATE"


# Do not have INTERVAL
class INTERVAL(sqltypes.Interval):
    __visit_name__ = "INTERVAL"


class VARCHAR(sqltypes.VARCHAR):
    __visit_name__ = "VARCHAR"


class VARBINARY(sqltypes.VARBINARY):
    __visit_name__ = "VARBINARY"


class CHAR(sqltypes.CHAR):
    __visit_name__ = "CHAR"



import asyncio
import sys

import aiordp


async def start(engine):
    async with engine.acquire() as conn:

        async for r in await conn.execute("select * from region"):
            print(r)


async def example(loop, host: str):

    engine = await aiordp.create_engine(
        host=host,
        port=4333,
        user="RAPIDS",
        password="rapids",
        db="sf1",
        loop=loop,
    )
    await start(engine)


def main(host: str):
    loop = asyncio.new_event_loop()
    print("Connect to", host)
    asyncio.run(example(loop, host=host))


if __name__ == "__main__":
    host = sys.argv[1] if len(sys.argv) > 1 else "localhost"
    main(host)



#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class StatementReceipt(object):
    """
    Attributes:
     - statementString
     - statementId
     - statementKey
    """


    def __init__(self, statementString=None, statementId=None, statementKey=None,):
        self.statementString = statementString
        self.statementId = statementId
        self.statementKey = statementKey

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.statementString = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I64:
                    self.statementId = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I64:
                    self.statementKey = iprot.readI64()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('StatementReceipt')
        if self.statementString is not None:
            oprot.writeFieldBegin('statementString', TType.STRING, 1)
            oprot.writeString(self.statementString.encode('utf-8') if sys.version_info[0] == 2 else self.statementString)
            oprot.writeFieldEnd()
        if self.statementId is not None:
            oprot.writeFieldBegin('statementId', TType.I64, 2)
            oprot.writeI64(self.statementId)
            oprot.writeFieldEnd()
        if self.statementKey is not None:
            oprot.writeFieldBegin('statementKey', TType.I64, 3)
            oprot.writeI64(self.statementKey)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.statementId is None:
            raise TProtocolException(message='Required field statementId is unset!')
        if self.statementKey is None:
            raise TProtocolException(message='Required field statementKey is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class QueryReceiptResponse(object):
    """
    Attributes:
     - statementReceipts
    """


    def __init__(self, statementReceipts=None,):
        self.statementReceipts = statementReceipts

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.statementReceipts = []
                    (_etype3, _size0) = iprot.readListBegin()
                    for _i4 in range(_size0):
                        _elem5 = StatementReceipt()
                        _elem5.read(iprot)
                        self.statementReceipts.append(_elem5)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('QueryReceiptResponse')
        if self.statementReceipts is not None:
            oprot.writeFieldBegin('statementReceipts', TType.LIST, 1)
            oprot.writeListBegin(TType.STRUCT, len(self.statementReceipts))
            for iter6 in self.statementReceipts:
                iter6.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.statementReceipts is None:
            raise TProtocolException(message='Required field statementReceipts is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(StatementReceipt)
StatementReceipt.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'statementString', 'UTF8', None, ),  # 1
    (2, TType.I64, 'statementId', None, None, ),  # 2
    (3, TType.I64, 'statementKey', None, None, ),  # 3
)
all_structs.append(QueryReceiptResponse)
QueryReceiptResponse.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'statementReceipts', (TType.STRUCT, [StatementReceipt, None], False), None, ),  # 1
)
fix_spec(all_structs)
del all_structs



# -*- coding: utf-8 -*-
import datetime
import errno
from decimal import Decimal
from typing import TYPE_CHECKING, Optional
import logging

from thrift import TSerialization
from thrift.protocol import TCompactProtocol

from pyrdpdb.messages.compatibility.ProtocolCompatibleRequest.ttypes import (
    ProtocolCompatibleRequest,
)
from pyrdpdb.messages.compatibility.ProtocolCompatibleResponse.ttypes import (
    ProtocolCompatibleResponse,
)
from pyrdpdb.messages.compatibility.ProtocolIncompatibleResponse.ttypes import (
    ProtocolIncompatibleResponse,
)
from pyrdpdb.messages.v4.AuthenticationOkResponse.ttypes import AuthenticationOkResponse
from pyrdpdb.messages.v4.ClearTextTunnelRequest.ttypes import ClearTextTunnelRequest
from pyrdpdb.messages.v4.ClosePreparedStatementRequest.ttypes import (
    ClosePreparedStatementRequest,
)
from pyrdpdb.messages.v4.ClosePreparedStatementResponse.ttypes import (
    ClosePreparedStatementResponse,
)
from pyrdpdb.messages.v4.ColumnMetadata.ttypes import DataType
from pyrdpdb.messages.v4.DisconnectNotification.ttypes import DisconnectNotification
from pyrdpdb.messages.v4.DisconnectRequest.ttypes import DisconnectRequest
from pyrdpdb.messages.v4.ErrorResponse.ttypes import ErrorResponse
from pyrdpdb.messages.v4.GssNegotiationMsg.ttypes import GssNegotiationMsg
from pyrdpdb.messages.v4.KerberosAuthenticationRequest.ttypes import (
    KerberosAuthenticationRequest,
)
from pyrdpdb.messages.v4.MessageTypes.ttypes import MessageTypes
from pyrdpdb.messages.v4.ParseStatementRequest.ttypes import ParseStatementRequest
from pyrdpdb.messages.v4.ParseStatementResponse.ttypes import ParseStatementResponse
from pyrdpdb.messages.v4.PlainAuthenticationRequest.ttypes import (
    PlainAuthenticationRequest,
)
from pyrdpdb.messages.v4.QueryExecuteRequest.ttypes import QueryExecuteRequest
from pyrdpdb.messages.v4.QueryExecuteResponse.ttypes import QueryExecuteResponse
from pyrdpdb.messages.v4.QueryProgressRequest.ttypes import QueryProgressRequest
from pyrdpdb.messages.v4.QueryProgressResponse.ttypes import QueryProgressResponse
from pyrdpdb.messages.v4.QueryReceiptResponse.ttypes import (
    QueryReceiptResponse,
    StatementReceipt,
)
from pyrdpdb.messages.v4.RowDataResponse.ttypes import RowDataResponse
from pyrdpdb.messages.v4.RowMetadataResponse.ttypes import RowMetadataResponse
from pyrdpdb.messages.v4.SessionChangeRequest.ttypes import SessionChangeRequest
from pyrdpdb.messages.v4.SessionChangeResponse.ttypes import SessionChangeResponse
from pyrdpdb.messages.v4.SessionDataResponse.ttypes import SessionDataResponse
from pyrdpdb.messages.v4.SessionDroppedMessagesResponse.ttypes import (
    SessionDroppedMessagesResponse,
)
from pyrdpdb.messages.v4.SessionMonitoringRequest.ttypes import SessionMonitoringRequest
from pyrdpdb.messages.v4.SessionMonitoringResponse.ttypes import SessionMonitoringResponse
from pyrdpdb.messages.v4.SessionProtocolVersionResponse.ttypes import (
    SessionProtocolVersionResponse,
)
from pyrdpdb.messages.v4.ShutdownNotification.ttypes import ShutdownNotification
from pyrdpdb.messages.v4.StatementBindAndExecuteRequest.ttypes import (
    StatementBindAndExecuteRequest,
)
from pyrdpdb.messages.v4.StatementResponse.ttypes import StatementResponse
from pyrdpdb.messages.v4.TunnelSupportedResponse.ttypes import TunnelSupportedResponse
from pyrdpdb.messages.v4.WarningResponse.ttypes import WarningResponse

from . import err
from .constants import CR, FIELD_TYPE
from .utils import byte2int, bytes2int, int2_4bytes, int2byte

# from .messages.v2.AuthenticationRequest.ttypes import AuthenticationRequest
# from .messages.v3.QueryCancellationRequest.ttypes import QueryCancellationRequest
# from .messages.v3.QueryCancellationResponse.ttypes import QueryCancellationResponse

log = logging.getLogger("pyrdp.wp")


# def test_log():
#     log.info("This is a test log message")
#     log.debug("This is a debug log message")
#     log.error("This is a error log message")


if TYPE_CHECKING:
    from socket import socket

typecode_map = {
    "BOOLEAN": FIELD_TYPE.TINY,
    "INTEGER": FIELD_TYPE.LONG,
    "BIGINT": FIELD_TYPE.LONG,
    "DECIMAL": FIELD_TYPE.DECIMAL,
    "FLOAT": FIELD_TYPE.FLOAT,
    "DATE": FIELD_TYPE.DATE,
    "VARCHAR": FIELD_TYPE.VARCHAR,
    # "VARBINARY": FIELD_TYPE.VARBINARY,
    "TIMESTAMP": FIELD_TYPE.TIMESTAMP,
}


class WirelineProtocol(object):

    MAX_MSG_PAYLOAD_SIZE = 16777208
    # 24 bits - 8 bytes for the header

    def __init__(self, connection=None, tprotocol=None):
        self.connection = connection
        self.sock: socket = connection._sock
        self.rfile = connection._rfile
        self.serialize = TSerialization.serialize
        self.deserialize = TSerialization.deserialize
        self.tprotocol = tprotocol
        self.current_row_data_response: Optional[RowDataResponse] = None
        self.new_row_data_response: Optional[RowDataResponse] = None
        self.has_row_metadata_response = False
        self.current_message_type = None  # current message type
        self.rdb_header = None
        self.icurrent_row_index = 0  # current row index
        self.imax_rows = 0  # total rows of this rowdataresponse packet
        self.row = None  # current row data
        self.numColumns = None  # number of columns
        self.columns = None  # columns metadata
        self.description = None
        self.end_of_response = False  # response end of this request
        self.end_of_results = False  # results end of this request
        self.errstr = None  # ErrorResponse message here
        self.statements: dict = {}
        self.finished_statements: list = []
        self.has_next: bool = False
        self.chunk: Optional[list] = None
        self.chunk_count = 0

        if tprotocol is None:
            self.tprotocol = TCompactProtocol.TCompactProtocolFactory()

    def login(self):
        """
        Before real request and response,  make sure server side have ability to
        response the request.
        """
        self._do_protocol_compatibility_dailog("login")
        self._do_tunnel_dialog()
        # Authentication
        if self.connection.auth_type == "PLAIN":
            self._do_plain_authentication_dialog()
        else:
            raise err.NotSupportedError("Only support Authentication type PLAIN")

        self._do_session_change_dialog()

    def _do_protocol_compatibility_dailog(self, desc):
        """
        Negotiate protocol version
        """
        request = ProtocolCompatibleRequest(4, desc)
        self._send_message(request, 1)
        message = self._read_message()
        return message

    def _do_tunnel_dialog(self):
        """
        Negotiate tunnel type
        """
        request = ClearTextTunnelRequest()
        self._send_message(request, 21)
        message = self._read_message()
        return message

    def _do_kerberos_authentication_dialog(self):
        """
        Authentication with kerberos, still not support here.
        """
        pass

    def _do_plain_authentication_dialog(self):
        """
        Authentication with username and password.
        """
        username = self.connection.user
        password = self.connection.password
        request = PlainAuthenticationRequest(username, password)
        self._send_message(request, 31)
        message = self._read_message()

        if message.type == 6:
            log.error("Could not authenticate with RapidsDB (username/password).")
            # throw a exception
            raise err.OperationalError(message.message.errorString, "28R01")

    # def _do_authentication_dialog(self, username):
    #     """
    #     Authentication, actually server side always allow client pass the authentication
    #     right now.
    #     """
    #     request = AuthenticationRequest(username)
    #     self._send_message(request, 23)
    #     message = self._read_message()
    #     return message

    def _do_session_change_dialog(self):
        """
        I have no idea about this method, seems we need this request befor real
        comunication.
        """
        request = SessionChangeRequest()
        self._send_message(request, 41)
        message = self._read_message()
        return message

    def query(self, sql):
        """
        Send query message and handle response.
        """
        self._clear()
        self.current_message_type = MessageTypes.QueryExecuteRequest

        request = QueryExecuteRequest(query=sql)
        self._send_message(request, 51)

        self._process_state_transitions()

        if self.new_row_data_response is not None:
            self.current_row_data_response = self.new_row_data_response
            self.imax_rows = len(self.new_row_data_response.rows)

    def _read_message(self):
        """
        Read a message from sever side. First read a message header , according to the
        header info, the read the total message.
        """
        header = self._read_bytes(8)
        length = bytes2int(header[0:4])[0]
        return_type = byte2int(header[4])
        payload_size = length - 8
        payload = self._read_bytes(payload_size)
        message = self._create_message(return_type)
        return_message = self.deserialize(message, payload, self.tprotocol)

        msgTypeName = self._get_messagetype_name(return_type)
        msg = (
            "Receving Message type: "
            + str(return_type)
            + " ("
            + msgTypeName
            + ") from server."
        )

        log.debug(msg)
        return MessageAndTypePair(return_message, return_type)

    def _read_bytes(self, num_bytes):
        """
        Read bytes from socket.
        """
        while True:
            try:
                data = self.rfile.read(num_bytes)
                break
            except (IOError, OSError) as e:
                if e.errno == errno.EINTR:
                    continue
                raise err.OperationalError(
                    CR.CR_SERVER_LOST,
                    "Lost connection to RapidDB server during query (%s)" % (e,),
                )
        if len(data) < num_bytes:
            raise err.OperationalError(
                CR.CR_SERVER_LOST,
                "Lost connection to RapidDB server during query ",
            )
        return data

    def _send_message(self, message, type):
        """
        Send  message to server side.
        """

        msgTypeName = self._get_messagetype_name(type)
        msg = f"Sending Message type: {str(type)} ({msgTypeName}) to server. "

        # log message info
        log.debug(msg)

        message_bytes = TSerialization.serialize(message, self.tprotocol)
        if len(message_bytes) > self.MAX_MSG_PAYLOAD_SIZE:
            raise err.ProgrammingError("Message is over max message payload size.")

        header = self._create_header(type, len(message_bytes))
        data = header + message_bytes
        self._send_bytes(data)

    def _send_bytes(self, data):
        """
        Send bytes to socket.
        """

        try:
            self.sock.sendall(data)
        except IOError as e:
            raise err.OperationalError(
                CR.CR_SERVER_GONE_ERROR,
                "RapidDB server has gone away (%r)" % (e,),
            )

    def _get_messagetype_name(self, type):
        """
        Get message type name according type id.
        """
        message_type_name = MessageTypes._VALUES_TO_NAMES[type]
        if message_type_name is None:
            raise err.ProgrammingError("Message type not exist.")
        return message_type_name

    def _query_metadata_response_handler(self, meta_message: RowMetadataResponse):
        self.has_next = True
        self.numColumns = meta_message.numColumns
        column_metadatas = meta_message.columns
        columns = []
        descs = []
        # sqlAlchemly only support list description
        for column_metadata in column_metadatas:
            col_info = {}
            desc = []
            col_info["name"] = column_metadata.columnName
            col_info["type"] = column_metadata.columnType
            col_info["rdb_type"] = self._get_rdb_type(column_metadata.columnType)
            col_info["precision"] = column_metadata.precision
            col_info["scale"] = column_metadata.scale
            col_info["nullable"] = column_metadata.isNullable
            # change 'type' to DBAPI 2.0 compliant type code
            col_info["type"] = typecode_map.get(col_info["rdb_type"].upper()) or 0
            desc.append(col_info["name"])
            desc.append(col_info["type"])
            desc.append(col_info["rdb_type"])
            desc.append(col_info["precision"])
            desc.append(col_info["scale"])
            desc.append(col_info["nullable"])
            columns.append(col_info)
            descs.append(tuple(desc))
        self.columns = columns
        self.description = tuple(descs)
        self.has_row_metadata_response = True
        self.imax_rows = 0
        self.icurrent_row_index = 0
        if self.numColumns > 0:
            self.end_of_response = False
            self.end_of_results = False

    def new_state_error_response_handler(self, response_pair) -> bool:
        # - errorCode
        # - sqlStateCode
        # - errorString
        # - stackTrace
        self._error_received = True
        message = response_pair.message
        self.errstr = message.errorString
        if message.stackTrace:
            self.errstr += "\n" + message.stackTrace
        b_done = True
        self.end_of_response = True
        return b_done

    def query_execute_request_handler(self, new_state, response_pair) -> bool:
        b_done = False
        if new_state == MessageTypes.QueryReceiptResponse:
            _receipt: QueryReceiptResponse = response_pair.message
            # self.statements = _receipt.statementReceipts
            stmt_receipt: StatementReceipt
            for stmt_receipt in _receipt.statementReceipts:
                self.statements[stmt_receipt.statementId] = {
                    "stmt": stmt_receipt.statementString,
                    "key": stmt_receipt.statementKey,
                }
        elif new_state == MessageTypes.WarningResponse:
            pass
        elif new_state == MessageTypes.ErrorResponse:
            b_done = self.new_state_error_response_handler(response_pair)
        else:
            # raise a execption to tell wrong state
            pass
        return b_done

    def query_receipt_response_handler(self, new_state, response_pair) -> bool:
        b_done = False
        if new_state == MessageTypes.RowMetadataResponse:
            self._query_metadata_response_handler(response_pair.message)
        elif new_state == MessageTypes.WarningResponse:
            pass
        elif new_state == MessageTypes.ErrorResponse:
            b_done = self.new_state_error_response_handler(response_pair)
        elif new_state == MessageTypes.StatementResponse:
            pass
        else:
            # TODO raise a execption to tell wrong state
            pass
        return b_done

    def row_metadata_response_handler(self, new_state, response_pair) -> bool:
        b_done = False
        if new_state == MessageTypes.RowDataResponse:
            row_message: RowDataResponse = response_pair.message
            # self.icurrent_row_index = 0
            self.imax_rows = len(row_message.rows)
            self.new_row_data_response = row_message
            self.end_of_response = False
        elif new_state == MessageTypes.StatementResponse:
            # stmt_message = response_pair.message
            self.imax_rows = 0
        elif new_state == MessageTypes.WarningResponse:
            pass
        elif new_state == MessageTypes.ErrorResponse:
            b_done = self.new_state_error_response_handler(response_pair)
        else:
            # TODO raise a execption to tell wrong state
            pass
        return b_done

    def row_data_response_handler(self, new_state, response_pair, b_done) -> bool:
        if new_state == MessageTypes.RowDataResponse:
            self.end_of_response = False
            b_done = True
        elif new_state == MessageTypes.StatementResponse:
            """
            statementId
            statementString
            numMatchedRows
            numAffectedRows
            affectedVerb
            """
            stmt_message: StatementResponse = response_pair.message
            self.affected_rows = stmt_message.numAffectedRows
            self.finished_statements.append(stmt_message.statementId)
            if set(self.finished_statements) == set(self.statements.keys()):
                self.has_next = False

        elif new_state == MessageTypes.WarningResponse:
            pass
        elif new_state == MessageTypes.ErrorResponse:
            b_done = self.new_state_error_response_handler(response_pair)
        else:
            # TODO raise a execption to tell wrong state
            pass
        return b_done

    def statement_response_handler(self, new_state, response_pair) -> bool:
        b_done = False
        if new_state == MessageTypes.QueryExecuteResponse:
            b_done = True
            self.end_of_response = True
            self.has_next = False
        elif new_state == MessageTypes.WarningResponse:
            pass
        elif new_state == MessageTypes.ErrorResponse:
            b_done = self.new_state_error_response_handler(response_pair)
        # TODO: Handle receiving another StatementResponse here
        # TODO: e.g., as part of a batch submitted as one query.
        elif new_state == MessageTypes.RowMetadataResponse:
            self._query_metadata_response_handler(response_pair.message)
        # do not handle multiple results
        elif new_state == MessageTypes.StatementResponse:
            pass
        else:
            # TODO raise a execption to tell wrong state
            self.has_next = False
        return b_done

    def query_execute_response_handler(self, new_state, response_pair) -> bool:
        # self.new_row_data_response = None
        # self.current_row_data_response = None
        # self.icurrent_row_index = 0
        # self.imax_rows = 0
        # self.numColumns = 0
        # self.columns = None
        # self.description = None
        # self.errstr = None
        # self.finished_statements = []
        # self.statements = {}
        # self.end_of_response = True
        # self.end_of_results = True
        # return True
        pass

    def error_response_handler(self, new_state, response_pair) -> bool:
        b_done = False
        if new_state == MessageTypes.QueryExecuteResponse:
            b_done = True
            self.end_of_response = True
        elif new_state == MessageTypes.WarningResponse:
            pass
        elif new_state == MessageTypes.ErrorResponse:
            b_done = self.new_state_error_response_handler(response_pair)
        else:
            # TODO raise a execption to tell wrong state
            pass
        return b_done

    def _handle_message_state(self):
        """Handle response from database server side."""
        b_done = False
        response_pair = None
        new_state = None

        new_state = self._read_rdp_header()
        log.debug(
            "Receive Message type: {} ({}) from server.".format(
                str(new_state), self._get_messagetype_name(new_state)
            )
        )

        if (
            self.current_message_type == MessageTypes.RowDataResponse
            and new_state == MessageTypes.RowDataResponse
        ):
            b_done = True
        else:
            response_pair = self._read_rdp_buffer()

        # log.debug(f"message: {response_pair.message}")

        if self.current_message_type == MessageTypes.QueryExecuteRequest:
            b_done = self.query_execute_request_handler(new_state, response_pair)
        elif self.current_message_type == MessageTypes.QueryReceiptResponse:
            b_done = self.query_receipt_response_handler(new_state, response_pair)
        elif self.current_message_type == MessageTypes.RowMetadataResponse:
            b_done = self.row_metadata_response_handler(new_state, response_pair)
        elif self.current_message_type == MessageTypes.RowDataResponse:
            b_done = self.row_data_response_handler(new_state, response_pair, b_done)
        elif self.current_message_type == MessageTypes.StatementResponse:
            b_done = self.statement_response_handler(new_state, response_pair)
        elif self.current_message_type == MessageTypes.QueryExecuteResponse:
            b_done = self.query_execute_response_handler(new_state, response_pair)
        elif self.current_message_type == MessageTypes.WarningResponse:
            # All warnings are handled in the context of other queries, and we ignore
            # warnings from the context
            # of setting current_message_type too. So we should never get here.
            pass
        elif self.current_message_type == MessageTypes.ErrorResponse:
            b_done = self.error_response_handler(new_state, response_pair)
        else:
            b_done = True

        self.rdb_header.msg_prior_type = self.current_message_type
        self.current_message_type = new_state

        return b_done

    # TODO: extracte repeat part
    def _handle_error_response(self):
        pass

    def _get_rdb_type(self, colum_type):
        """
        Get pyrdpdb data type according to metadata colum type.
        :param number colum_type : colum data type.
        : return : pyrdpdb data type.
        """
        return DataType._VALUES_TO_NAMES[colum_type]

    def has_next_row(self) -> bool:
        return self.icurrent_row_index < self.imax_rows

    def get_next_row(self):
        """
        Get next row data.
        : return : flag of if have next row.
        """
        if (
            self.end_of_response or self.end_of_results
        ) and self.current_row_data_response is None:
            return False
        if self.icurrent_row_index < self.imax_rows:
            self._next_row()
        else:
            read_data = self._read_more_data()
            if read_data:
                self.icurrent_row_index = 0
                self.end_of_results = False
                self._next_row()
            else:
                self.end_of_response = True
                self.end_of_results = True
        if self.end_of_results:
            return False
        return True

    def to_pandas(self) -> bool:
        """Generate self._df dataframe from data batch."""

        if (
            self.end_of_response or self.end_of_results
        ) and self.current_row_data_response is None:
            return False

        self._df = None

        if self.icurrent_row_index < self.imax_rows:
            df = self._next_set_column()
        else:
            more_data = self._read_more_data()
            if more_data:
                self.icurrent_row_index = 0
                self.end_of_results = False
                df = self._next_set_column()
            else:
                self.end_of_response = True
                self.end_of_results = True
        if self.end_of_results:
            return False
        self._df = df
        return True

    def _next_set_column(self):
        """
        Decode all buffered data to a set.
        """
        import pandas as pd

        if self.end_of_response and self.current_row_data_response is None:
            return

        self.chunk = []
        columns = [desc[0] for desc in self.description]
        column_dict = {col: [] for col in columns}
        while self.icurrent_row_index < self.imax_rows:
            row = self.current_row_data_response.rows[self.icurrent_row_index]
            self.icurrent_row_index += 1
            row_columns = row.columns
            for col_idx, value in enumerate(row_columns):
                col_name = columns[col_idx]
                col_type = self.columns[col_idx]["rdb_type"]
                value = row_columns[col_idx]
                _new_value = self.convert_db_value(value, col_type)
                column_dict[col_name].append(_new_value)

        df = pd.DataFrame(column_dict)
        return df

    def convert_db_value(self, value, rdb_type):
        if value is None:
            return None
        elif rdb_type == "BOOLEAN":
            return value.booleanVal
        elif rdb_type == "TINYINT":
            return value.tinyIntVal
        elif rdb_type == "SMALLINT":
            return value.smallIntVal
        elif rdb_type == "INTEGER":
            return value.integerVal
        elif rdb_type == "BIGINT":
            return value.bigintVal
        elif rdb_type == "DECIMAL":
            return Decimal(value.bigDecimalVal)
        elif rdb_type == "FLOAT":
            return value.doubleVal
        elif rdb_type == "DATE":
            ts = value.dateVal
            return datetime.date(year=ts.year, month=ts.month, day=ts.day)
        elif rdb_type == "TIME":
            ts = value.timeVal
            return datetime.time(
                hour=ts.hour,
                minute=ts.minute,
                second=ts.second,
                microsecond=int(ts.nanoseconds / 1000),
            )
        elif rdb_type == "TIMESTAMP":
            ts = value.timestampVal
            return datetime.datetime(
                ts.year,
                ts.month,
                ts.day,
                ts.hour,
                ts.minute,
                ts.second,
                int(ts.nanoseconds / 1000),
            )
        elif rdb_type == "INTERVAL":
            return value.intervalVal
        elif rdb_type == "VARCHAR":
            return value.stringVal
        elif rdb_type == "VARBINARY":
            return value.binaryVal
        else:
            raise ValueError(f"Unsupported data type: {rdb_type}")

    def get_next_set(self):
        """
        Get next set data.
        """
        if (
            self.end_of_response or self.end_of_results
        ) and self.current_row_data_response is None:
            self.row = None
            self.chunk = None
            return False

        if self.icurrent_row_index < self.imax_rows:
            self._next_set()
        else:
            more_data = self._read_more_data()
            if more_data:
                self.icurrent_row_index = 0
                self.end_of_results = False
                self._next_set()
                log.debug(f"chunk size: {len(self.chunk)}")
            else:
                self.end_of_response = True
                self.end_of_results = True
        if self.end_of_results and self.end_of_response:
            return False
        return True

    def _next_set(self):
        """
        Decode all buffered data to a set.
        """
        if self.end_of_response and self.current_row_data_response is None:
            return

        self.chunk = []
        while not self._get_row_from_rdp_buffer():
            self.chunk.append(self.row)

    def _next_row(self):
        """
        Decode next row from buffer.
        """
        if self.end_of_response and self.current_row_data_response is None:
            return

        self._get_row_from_rdp_buffer()

    def _get_row_from_rdp_buffer(self) -> bool:
        """Run row protocol conversion and return True if there is no more records
        in buffer and False otherwise.

        Note: each time one row is read from buffer, convert column data types
        from RDP to python, save row to self.row.
        """
        if (
            self.current_row_data_response is not None
            and self.icurrent_row_index < self.imax_rows
        ):
            row = self.current_row_data_response.rows[self.icurrent_row_index]
            self.icurrent_row_index += 1
            columns = row.columns
            temp_row = []
            for i in range(len(columns)):
                if columns[i].isNull:
                    temp_row.append(None)
                elif self.columns[i]["rdb_type"] == "NULL":
                    temp_row.append(None)
                elif self.columns[i]["rdb_type"] == "BOOLEAN":
                    temp_row.append(columns[i].booleanVal)
                elif self.columns[i]["rdb_type"] == "TINYINT":
                    temp_row.append(columns[i].tinyIntVal)
                elif self.columns[i]["rdb_type"] == "SMALLINT":
                    temp_row.append(columns[i].smallIntVal)
                elif self.columns[i]["rdb_type"] == "INTEGER":
                    temp_row.append(columns[i].integerVal)
                elif self.columns[i]["rdb_type"] == "BIGINT":
                    temp_row.append(columns[i].bigintVal)
                elif self.columns[i]["rdb_type"] == "DECIMAL":
                    temp_row.append(Decimal(columns[i].bigDecimalVal))
                    # temp_row.append(colums[i].bigDecimalVal)
                elif self.columns[i]["rdb_type"] == "FLOAT":
                    temp_row.append(columns[i].doubleVal)
                elif self.columns[i]["rdb_type"] == "DATE":
                    ts = columns[i].dateVal
                    date_val = datetime.date(year=ts.year, month=ts.month, day=ts.day)
                    temp_row.append(date_val)
                elif self.columns[i]["rdb_type"] == "TIME":
                    ts = columns[i].timeVal
                    time_val = datetime.time(
                        hour=ts.hour,
                        minute=ts.minute,
                        second=ts.second,
                        microsecond=int(ts.nanoseconds / 1000),
                    )
                    temp_row.append(time_val)
                elif self.columns[i]["rdb_type"] == "TIMESTAMP":
                    # ts type: ..messages.v4.ColumnValue.ttypes.Timestamp
                    ts = columns[i].timestampVal
                    dt = datetime.datetime(
                        ts.year,
                        ts.month,
                        ts.day,
                        ts.hour,
                        ts.minute,
                        ts.second,
                        int(ts.nanoseconds / 1000),  # to microsecond
                    )
                    temp_row.append(dt)
                elif self.columns[i]["rdb_type"] == "INTERVAL":
                    interval = columns[i].intervalVal
                    temp_row.append(interval)
                elif self.columns[i]["rdb_type"] == "VARCHAR":
                    temp_row.append(columns[i].stringVal)
                elif self.columns[i]["rdb_type"] == "VARBINARY":
                    temp_row.append(columns[i].binaryVal)
                else:
                    raise err.NotSupportedError(
                        "Unimplemented data type from the wire line protocol: {}".format(
                            self.columns[i]["rdb_type"]
                        )
                    )
            self.row = tuple(temp_row)
            self.end_of_results = False
            return False
        else:
            self.end_of_results = True
            return True

    def _read_more_data(self) -> bool:
        """
        Read more data.
        : return : flag of the response state.
        """
        self.chunk_count += 1
        log.debug(f"Reading batch {self.chunk_count}.")

        b_done = False
        self.current_row_data_response = None

        if self.end_of_response:
            return False
        response_pair = None
        prior_state = self.rdb_header.msg_prior_type
        new_state = self.rdb_header.msg_type

        if (
            prior_state == MessageTypes.RowDataResponse
            and new_state == MessageTypes.RowDataResponse
            and self.end_of_response is False
        ):
            response_pair = self._read_rdp_buffer()
            row_data_response: RowDataResponse = response_pair.message
            self.imax_rows = len(row_data_response.rows)
            self.current_row_data_response = row_data_response
            self.icurrent_row_index = 0
            b_done = True
        else:
            return False

        self.current_message_type = new_state
        self._process_state_transitions()

        return b_done

    def _process_state_transitions(self):
        """Process Thrift server message state transitions until all message responses
        are finished."""
        done = False
        while not done:
            done = self._handle_message_state()

    def _read_rdp_header(self):
        """
        Read 8 bytes from socket and resolv the message meta information.
        : return : message type id.
        """
        header = self._read_bytes(8)
        length = bytes2int(header[0:4])[0]
        return_type = byte2int(header[4])
        payload_size = length - 8

        if self.rdb_header is None:
            self.rdb_header = RDPHeader()
        self.rdb_header.raw_msg_type = return_type
        self.rdb_header.msg_type = return_type
        self.rdb_header.msg_length = payload_size
        self.rdb_header.clear()

        return return_type

    def _read_rdp_buffer(self):
        """
        Read message boay according to message payload size.
        : return : MessageAndTypePair.
        """
        payload_size = self.rdb_header.msg_length
        payload = self._read_bytes(payload_size)
        message = self._create_message(self.rdb_header.raw_msg_type)
        self.deserialize(message, payload, self.tprotocol)
        return MessageAndTypePair(message, self.rdb_header.raw_msg_type)

    def _create_header(self, type, size):
        """
        Create byte mesage header.
        :param number type : message type id.
        :param number size : message paload size.
        : return : bytes of message header.
        """
        header = (
            int2_4bytes(size + 8)
            + int2byte(type)
            + int2byte(0)
            + int2byte(0)
            + int2byte(0)
        )
        return header

    def _create_message(self, type):
        """
        Create empty message according to message type id.
        :param number type : message type id.
        : return : Empty message of the type id.
        """

        if type == 1:
            return ProtocolCompatibleRequest()
        elif type == 2:
            return ProtocolCompatibleResponse()
        elif type == 4:
            return ProtocolIncompatibleResponse()
        elif type == 6:
            return ErrorResponse()
        elif type == 8:
            return WarningResponse()
        elif type == 10:
            return ShutdownNotification()
        elif type == 12:
            return DisconnectNotification()
        elif type == 13:
            return DisconnectRequest()
        elif type == 15:
            return DisconnectRequest()
        elif type == 16:
            return DisconnectRequest()
        elif type == 21:
            return ClearTextTunnelRequest()
        elif type == 22:
            return TunnelSupportedResponse()
        elif type == 31:
            return PlainAuthenticationRequest()
        elif type == 33:
            return KerberosAuthenticationRequest()
        elif type == 34:
            return GssNegotiationMsg()
        elif type == 38:
            return AuthenticationOkResponse()
        elif type == 41:
            return SessionChangeRequest()
        elif type == 42:
            return SessionChangeResponse()
        elif type == 51:
            return QueryExecuteRequest()
        elif type == 52:
            return QueryReceiptResponse()
        elif type == 54:
            return RowMetadataResponse()
        elif type == 56:
            return RowDataResponse()
        elif type == 58:
            return StatementResponse()
        elif type == 60:
            return QueryExecuteResponse()
        # elif type == 71:
        #     return QueryCancellationRequest()
        # elif type == 72:
        #     return QueryCancellationResponse()
        elif type == 81:
            return QueryProgressRequest()
        elif type == 82:
            return QueryProgressResponse()
        elif type == 91:
            return SessionMonitoringRequest()
        elif type == 92:
            return SessionProtocolVersionResponse()
        elif type == 94:
            return SessionDataResponse()
        elif type == 96:
            return SessionDroppedMessagesResponse()
        elif type == 98:
            return SessionMonitoringResponse()
        elif type == 101:
            return ParseStatementRequest()
        elif type == 102:
            return ParseStatementResponse()
        elif type == 103:
            return StatementBindAndExecuteRequest()
        elif type == 105:
            return ClosePreparedStatementRequest()
        elif type == 106:
            return ClosePreparedStatementResponse()
        else:
            return None

    def _clear(self):
        self.errstr = None
        self.row = None
        self.numColumns = None
        self.columns = None
        self.description = None
        # self.end_of_response = False
        # self.end_of_results = False
        self.statements = {}
        self.finished_statements = []
        self.has_next = False
        self.chunk = None
        self.chunk_count = 0
        self.icurrent_row_index = 0
        self.imax_rows = 0
        # self.current_message_type = None
        # self.rdb_header = None
        self.current_row_data_response = None
        self.new_row_data_response = None
        # self.has_row_metadata_response = False


class RDPHeader(object):
    def __init__(
        self,
        message=None,
        msg_type=None,
        msg_prior_type=None,
        raw_msg_type=None,
        msg_length=None,
    ):
        self.message = message
        self.msg_type = msg_type
        self.msg_prior_type = msg_prior_type
        self.raw_msg_type = raw_msg_type
        self.msg_length = msg_length

    def clear(self):
        self.message = None


class MessageAndTypePair(object):
    """Message and message type wrapper."""

    def __init__(self, message, type):
        self.message = message
        self.type = type


class TableMetaData(object):
    def __init__(self):
        self.catalog = None
        self.schema = None
        self.name = None



import asyncio
import sys

import aiordp


async def test_example(loop, host: str):
    rdp = await aiordp.connect(
        host=host,
        port=4333,
        user="RAPIDS",
        password="rapids",
        database="moxe",
        loop=loop,
    )
    print("Using Async cursor:")
    async with rdp.cursor() as cursor:
        await cursor.execute("select 1")
        rset = await cursor.fetchall()
        for r in rset:
            print(r)


async def test_sscursor(loop, host: str):
    rdp = await aiordp.connect(
        host=host,
        port=4333,
        user="RAPIDS",
        password="rapids",
        database="sf1",
        cursorclass=aiordp.cursors.Cursor,
        loop=loop,
    )
    print("\nUsing Async SSCursor:\n")
    async with rdp.cursor() as cursor:
        await cursor.execute("select * from nation limit 10")
        rset = await cursor.fetchall()
        for r in rset:
            print(r)


def main(host: str):
    loop = asyncio.new_event_loop()
    print("Connect to", host)
    asyncio.run(test_example(loop, host))
    asyncio.run(test_sscursor(loop, host))


if __name__ == "__main__":
    host = sys.argv[1] if len(sys.argv) > 1 else "localhost"
    main(host)



import asyncio
import sys

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.sql import text

metadata = sa.MetaData()

tbl = sa.Table(
    "TBL",
    metadata,
    sa.Column("id", sa.Integer, primary_key=True),
    sa.Column("val", sa.String(255)),
    schema="MOXE",
)


async def create_table(engine):
    async with engine.begin() as conn:
        await conn.execute(text("DROP TABLE IF EXISTS MOXE.tbl"))
        await conn.execute(
            text(
                """CREATE TABLE MOXE.tbl (
                              id integer PRIMARY KEY,
                              val varchar(255))"""
            )
        )


async def go(host: str):
    engine = create_async_engine(f"rapidsdb+asyncrdp://RAPIDS:rapids@{host}:4333/moxe")
    await create_table(engine)
    async with engine.begin() as conn:
        await conn.execute(tbl.insert().values(id=1, val="abc"))
        await conn.execute(tbl.insert().values(id=2, val="xyz"))

        result = await conn.execute(tbl.select())
        for row in result:
            print(row[0], row[1])

        await conn.execute(text("drop table MOXE.tbl"))


def main(host: str):
    print("Connect to", host)
    asyncio.run(go(host))


if __name__ == "__main__":
    host = sys.argv[1] if len(sys.argv) > 1 else "localhost"
    main(host)



#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class MessageTypes(object):
    ProtocolCompatibleRequest = 1
    ProtocolCompatibleResponse = 2
    ProtocolIncompatibleResponse = 4
    ErrorResponse = 6
    WarningResponse = 8
    ShutdownNotification = 10
    DisconnectNotification = 12
    DisconnectRequest = 13
    PingRequest = 15
    PingResponse = 16
    ClearTextTunnelRequest = 21
    TunnelSupportedResponse = 22
    PlainAuthenticationRequest = 31
    KerberosAuthenticationRequest = 33
    GssNegotiationMsg = 34
    AuthenticationOkResponse = 38
    SessionChangeRequest = 41
    SessionChangeResponse = 42
    QueryExecuteRequest = 51
    QueryReceiptResponse = 52
    RowMetadataResponse = 54
    RowDataResponse = 56
    StatementResponse = 58
    QueryExecuteResponse = 60
    QueryProgressRequest = 81
    QueryProgressResponse = 82
    SessionMonitoringRequest = 91
    SessionProtocolVersionResponse = 92
    SessionDataResponse = 94
    SessionDroppedMessagesResponse = 96
    SessionMonitoringResponse = 98
    ParseStatementRequest = 101
    ParseStatementResponse = 102
    StatementBindAndExecuteRequest = 103
    ClosePreparedStatementRequest = 105
    ClosePreparedStatementResponse = 106

    _VALUES_TO_NAMES = {
        1: "ProtocolCompatibleRequest",
        2: "ProtocolCompatibleResponse",
        4: "ProtocolIncompatibleResponse",
        6: "ErrorResponse",
        8: "WarningResponse",
        10: "ShutdownNotification",
        12: "DisconnectNotification",
        13: "DisconnectRequest",
        15: "PingRequest",
        16: "PingResponse",
        21: "ClearTextTunnelRequest",
        22: "TunnelSupportedResponse",
        31: "PlainAuthenticationRequest",
        33: "KerberosAuthenticationRequest",
        34: "GssNegotiationMsg",
        38: "AuthenticationOkResponse",
        41: "SessionChangeRequest",
        42: "SessionChangeResponse",
        51: "QueryExecuteRequest",
        52: "QueryReceiptResponse",
        54: "RowMetadataResponse",
        56: "RowDataResponse",
        58: "StatementResponse",
        60: "QueryExecuteResponse",
        81: "QueryProgressRequest",
        82: "QueryProgressResponse",
        91: "SessionMonitoringRequest",
        92: "SessionProtocolVersionResponse",
        94: "SessionDataResponse",
        96: "SessionDroppedMessagesResponse",
        98: "SessionMonitoringResponse",
        101: "ParseStatementRequest",
        102: "ParseStatementResponse",
        103: "StatementBindAndExecuteRequest",
        105: "ClosePreparedStatementRequest",
        106: "ClosePreparedStatementResponse",
    }

    _NAMES_TO_VALUES = {
        "ProtocolCompatibleRequest": 1,
        "ProtocolCompatibleResponse": 2,
        "ProtocolIncompatibleResponse": 4,
        "ErrorResponse": 6,
        "WarningResponse": 8,
        "ShutdownNotification": 10,
        "DisconnectNotification": 12,
        "DisconnectRequest": 13,
        "PingRequest": 15,
        "PingResponse": 16,
        "ClearTextTunnelRequest": 21,
        "TunnelSupportedResponse": 22,
        "PlainAuthenticationRequest": 31,
        "KerberosAuthenticationRequest": 33,
        "GssNegotiationMsg": 34,
        "AuthenticationOkResponse": 38,
        "SessionChangeRequest": 41,
        "SessionChangeResponse": 42,
        "QueryExecuteRequest": 51,
        "QueryReceiptResponse": 52,
        "RowMetadataResponse": 54,
        "RowDataResponse": 56,
        "StatementResponse": 58,
        "QueryExecuteResponse": 60,
        "QueryProgressRequest": 81,
        "QueryProgressResponse": 82,
        "SessionMonitoringRequest": 91,
        "SessionProtocolVersionResponse": 92,
        "SessionDataResponse": 94,
        "SessionDroppedMessagesResponse": 96,
        "SessionMonitoringResponse": 98,
        "ParseStatementRequest": 101,
        "ParseStatementResponse": 102,
        "StatementBindAndExecuteRequest": 103,
        "ClosePreparedStatementRequest": 105,
        "ClosePreparedStatementResponse": 106,
    }
fix_spec(all_structs)
del all_structs



# mypy: ignore-errors

r"""
.. dialect:: rapidsdb+asyncrdp
    :name: asyncrdp
    :dbapi: asyncrdp
    :connectstring: rapidsdb+asyncrdp://user:password@host:port/dbname[?key=value&key=value...]

The asyncrdp dialect is SQLAlchemy's second Python asyncio dialect.

Using a special asyncio mediation layer, the asyncrdp dialect is usable
as the backend for the :ref:`SQLAlchemy asyncio <asyncio_toplevel>`
extension package.

This dialect should normally be used only with the
:func:`_asyncio.create_async_engine` engine creation function::

    from sqlalchemy.ext.asyncio import create_async_engine

    engine = create_async_engine(
        "rapidsdb+asyncrdp://user:pass@hostname:4333/federation/schema"
    )

"""  # noqa
from typing import TYPE_CHECKING

from sqlalchemy import pool, util
from sqlalchemy.connectors.asyncio import (
    AsyncAdapt_dbapi_connection,
    AsyncAdapt_dbapi_cursor,
    AsyncAdapt_dbapi_ss_cursor,
)
from sqlalchemy.dialects import registry
from sqlalchemy.pool.base import _ConnectionFairy
from sqlalchemy.util.concurrency import await_only

from .base import RDPDialect

if TYPE_CHECKING:
    from ..connections import Connection


class AsyncAdapt_asyncrdp_cursor(AsyncAdapt_dbapi_cursor):
    __slots__ = ()

    def _make_new_cursor(self, connection):
        return connection.cursor(self._adapt_connection.dbapi.Cursor)


class AsyncAdapt_asyncrdp_ss_cursor(
    AsyncAdapt_dbapi_ss_cursor, AsyncAdapt_asyncrdp_cursor
):
    __slots__ = ()

    def _make_new_cursor(self, connection):
        return connection.cursor(self._adapt_connection.dbapi.aiordp.cursors.SSCursor)


class AsyncAdapt_asyncrdp_connection(AsyncAdapt_dbapi_connection):
    __slots__ = ()

    _cursor_cls = AsyncAdapt_asyncrdp_cursor
    _ss_cursor_cls = AsyncAdapt_asyncrdp_ss_cursor

    # def ping(self, reconnect):
    #     assert not reconnect
    #     return await_(self._connection.ping(reconnect))

    def character_set_name(self):
        return self._connection.character_set_name()

    # def autocommit(self, value):
    #     await_(self._connection.autocommit(value))

    def terminate(self):
        # it's not awaitable.
        self._connection.close()

    def close(self) -> None:
        await_only(self._connection.ensure_closed())
        # await self._connection.ensure_closed()


class AsyncAdapt_asyncrdp_dbapi:
    def __init__(self, aiordp, pyrdp):
        self.aiordp = aiordp
        self.pyrdp = pyrdp
        self.paramstyle = "format"
        self._init_dbapi_attributes()
        self.Cursor, self.SSCursor = self._init_cursors_subclasses()

    def _init_dbapi_attributes(self):
        for name in (
            "Warning",
            "Error",
            "InterfaceError",
            "DataError",
            "DatabaseError",
            "OperationalError",
            "InterfaceError",
            "IntegrityError",
            "ProgrammingError",
            "InternalError",
            "NotSupportedError",
        ):
            setattr(self, name, getattr(self.aiordp, name))

        for name in (
            "NUMBER",
            "STRING",
            "DATETIME",
            "BINARY",
            "TIMESTAMP",
            "Binary",
        ):
            setattr(self, name, getattr(self.pyrdp, name))

    def connect(self, *arg, **kw):
        creator_fn = kw.pop("async_creator_fn", self.aiordp.connect)
        conn_ = AsyncAdapt_asyncrdp_connection(
            self,
            await_only(creator_fn(*arg, **kw)),
        )

        return conn_

    def _init_cursors_subclasses(self):
        # suppress unconditional warning emitted by asyncrdp
        class Cursor(self.aiordp.Cursor):
            async def _show_warnings(self, conn):
                pass

        class SSCursor(self.aiordp.SSCursor):
            async def _show_warnings(self, conn):
                pass

        return Cursor, SSCursor


class RDPDialect_asyncrdp(RDPDialect):
    driver = "asyncrdp"
    supports_statement_cache = True
    supports_savepoints = False
    supports_server_side_cursors = True
    _sscursor = AsyncAdapt_asyncrdp_ss_cursor

    is_async = True
    has_terminate = True

    @classmethod
    def import_dbapi(cls):
        try:
            from pyrdpdb import aiordp, pyrdp

            return AsyncAdapt_asyncrdp_dbapi(aiordp, pyrdp)
        except ImportError as e:
            raise ImportError(f"Failed to import required modules: {e}")

    @classmethod
    def get_pool_class(cls, url):
        async_fallback = url.query.get("async_fallback", False)

        if util.asbool(async_fallback):
            return pool.FallbackAsyncAdaptedQueuePool
        else:
            return pool.AsyncAdaptedQueuePool

    def do_terminate(self, dbapi_connection) -> None:
        """Terminate a DBAPI connection.

        Parameters:
        dbapi_connection: a :pep:`249` database connection.
        """
        dbapi_connection.terminate()

    def is_disconnect(self, e, connection: _ConnectionFairy, cursor):
        """Tells whether connection is closed.

        Parameters:
        - error: not used
        - connection: Optional[Connection]
        - cursor: not use

        Return: bool
        """
        conn: "Connection" = connection.connection._connection
        if conn is None:
            return True
        return conn.closed

    def get_driver_connection(self, connection):
        return connection


dialect = RDPDialect_asyncrdp
registry.register(
    "rapidsdb.asyncrdp", "pyrdpdb.aiordp.sa.asyncrapids", "RDPDialect_asyncrdp"
)



import asyncio
import sys

from pyrdpdb import aiordp, pyrdp

args = [(1, "a"), (2, "b"), (3, "c")]
ddl = "create table tbl (id integer, name varchar(50))"


def many_pyrdp() -> None:
    rdp = pyrdp.connect(
        host="localhost", port=4333, user="RAPIDS", password="rapids", db="moxe"
    )

    cursor: pyrdp.Cursor = rdp.cursor()

    print("create table tbl")
    if cursor.has_table("tbl"):
        cursor.execute("drop table tbl")
    cursor.execute(ddl)

    print("\nRun executemany() for insert")
    cursor.timing()  # turn on execution timing

    rcount = cursor.executemany("insert into tbl (id, name) VALUES (%s, %s)", args)

    print("\nInsert many count:", rcount)

    cursor.execute("select * from tbl")
    print("\nQuery result set:")
    print(cursor.fetchall())


async def many_aio(loop) -> None:
    rdp = await aiordp.connect(
        host="localhost",
        port=4333,
        user="RAPIDS",
        password="rapids",
        database="moxe",
        loop=loop,
    )
    cursor: aiordp.Cursor
    async with rdp.cursor() as cursor:
        if await cursor.has_table("tbl"):
            await cursor.execute("drop table tbl")
        await cursor.execute(ddl)

        print("\nRun async executemany() for insert")
        cursor.timing()  # turn on execution timing

        rcount = await cursor.executemany(
            "insert into tbl (id, name) VALUES (%s, %s)", args
        )
        print("\nAsync insert many count:", rcount)

        await cursor.execute("select * from tbl")
        print("\nQuery result set:")
        print(await cursor.fetchall())


def main(mode: str):
    print("Run mode:", mode)
    if mode == "pyrdp":
        many_pyrdp()
    else:
        loop = asyncio.new_event_loop()
        asyncio.run(many_aio(loop))


def usage():
    print("Usage: python many.py [pyrdp | aiordp]", file=sys.stderr)
    sys.exit(1)


if __name__ == "__main__":
    if len(sys.argv) != 2 or sys.argv[1] not in ["pyrdp", "aiordp"]:
        usage()
    main(sys.argv[1])



# flake8: noqa: E501
# -*- coding: utf-8 -*-
import re
from datetime import datetime
from itertools import count
from typing import TYPE_CHECKING, Any, Iterable, Optional, Union
import pandas as pd

from pyrdpdb import pyrdp

from . import err
from ._compat import range_type
from .utils import timeit
import logging

log = logging.getLogger("pyrdp.cursors")

RE_INSERT_VALUES = re.compile(
    r"\s*((?:INSERT|REPLACE)\b.+\bVALUES?\s*)(\(\s*(?:%s|%\(.+\)s)\s*(?:,\s*(?:%s|%\(.+\)s)\s*)*\))(\s*(?:ON DUPLICATE.*)?);?\s*\Z",
    re.IGNORECASE | re.DOTALL,
)


if TYPE_CHECKING:
    from .connections import Connection, ResultSet


class Cursor(object):
    """
    This is the object you use to interact with the database.
    Do not create an instance of a Cursor yourself. Call
    connections.Connection.cursor().
    See `Cursor <https://www.python.org/dev/peps/pep-0249/#cursor-objects>`_ in
    the specification.
    """

    Warning = err.Warning
    Error = err.Error
    InterfaceError = err.InterfaceError
    DatabaseError = err.DatabaseError
    DataError = err.DataError
    OperationalError = err.OperationalError
    IntegrityError = err.IntegrityError
    InternalError = err.InternalError
    ProgrammingError = err.ProgrammingError
    NotSupportedError = err.NotSupportedError
    RDBError = err.RDBError

    def __init__(self, connection: "Connection", echo=False):
        self.connection: Optional["Connection"] = connection
        self.description = None
        self.rownumber = 0
        # specifies the number of rows that the last .execute*() produced  or affected
        self.rowcount = -1
        # specifies the number of rows to fetch at a time with .fetchmany()
        self.arraysize = 1
        self._executed: Optional[str] = None
        self._result: Optional[ResultSet] = None  # response result of executed query
        self._rows = None  # all data rows
        self._timing = False
        # self._warnings_handled = False
        self._echo = echo

    def close(self):
        """
        Close cursor .
        """
        conn = self.connection
        if conn is None:
            return
        try:
            while self.nextset():
                pass
        except Exception:
            pass
        finally:
            self.connection = None

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        del exc_info
        self.close()

    def __iter__(self):
        return self

    def __next__(self):
        row = self.fetchone()
        if row is None:
            raise StopIteration
        return row

    def _get_db(self):
        if not self.connection:
            raise err.ProgrammingError("Cursor closed")
        return self.connection

    def execute(
        self,
        query,
        query_args=None,
        quiet: bool = True,
        internal: bool = False,
        generator: bool = False,
    ) -> int:
        """
        execute SQL statment.
        :param str query : query .
        :param tuple|dict|set query_args : query args.
        :return : affected row of this execute
        """
        if self.connection and self.connection._closed:
            raise pyrdp.Error("Already closed")
        aa = query.upper().replace(" ", "")
        if aa.startswith("CREATEINDEX"):
            return 0

        while self.nextset():
            pass

        stmt = self.mogrify(str(query), query_args)

        log.debug("generated query to run: " + stmt)
        self._query(stmt, generator=generator)
        self._executed = stmt

        if not quiet or self._timing and not internal:
            self._execute_stdout()

        return self.rowcount

    def _execute_stdout(self, start: Optional[datetime] = None):
        """Show rowcount and execution time if cursor.timing is True."""
        if self._timing:
            print(f"{self.rowcount} rows ({self._get_execute_time(start)} sec)")
        else:
            print(f"{self.rowcount} rows")

    def timing(self) -> None:
        """Flip show execution time flag setting."""
        self._timing = not self._timing
        print(f"Timing is {'ON' if self._timing else 'OFF'}.")

    def mogrify(self, query: str, args=None):
        """Returns real query that will be sent to database.
        :param str query : query .
        :param tuple|dict|set args : query args.
        :return : real sql .
        """
        if args is not None:
            query = query % self._escape_args(args)

        if not query.endswith(";"):
            query += ";"  # only required by RDP

        return query

    def _escape_args(self, query_args: Any) -> Any:
        """Formalize query by escaping all objects in query."""
        conn: "Connection" = self._get_conn()
        if isinstance(query_args, (tuple, list)):
            return tuple(conn.literal(arg) for arg in query_args)
        elif isinstance(query_args, dict):
            return {key: conn.literal(val) for (key, val) in query_args.items()}
        else:
            return conn.escape(query_args)

    def executemany(
        self,
        query: str,
        query_args: Iterable[Union[dict, list, tuple]],
        quiet: bool = True,
    ) -> int:
        """Execute query with sequence of arg for multiple times.

        :param query: query statement.
        :param args : query args.
        :return : affected row (rowcount) of this execute
        """
        if not query_args:
            return 0

        m = RE_INSERT_VALUES.match(query)
        if m:
            q_prefix = m.group(1) % ()
            q_values = m.group(2).rstrip()
            q_postfix = m.group(3) or ""
            assert q_values[0] == "(" and q_values[-1] == ")"
            return self._do_execute_many(q_prefix, q_values, q_postfix, query_args, quiet)

        start = datetime.now()
        self.rowcount = sum(self.execute(query, arg, quiet=True) for arg in query_args)

        if not quiet or self._timing:
            self._execute_stdout(start)

        return self.rowcount

    def _do_execute_many(
        self,
        prefix: str,
        placeholders: str,
        postfix: str,
        qargs: Iterable[Union[dict, list, tuple]],
        quiet: bool = True,
    ):
        sql = prefix
        start = datetime.now()
        num_placeholders = placeholders.count("%s")

        self._validate_args(qargs, num_placeholders)
        for arg in qargs:
            formatted_args = self._escape_args(arg)
            if num_placeholders < len(formatted_args):
                raise ValueError(
                    "Number of elements in values tuple does not match number of placeholders"
                )
            v = placeholders % formatted_args
            sql += v
            sql += ","
        sql = sql[:-1]  # remove the trailing ',' from generated query

        log.debug(f"[executemany] query: {sql}")
        self.execute(sql + postfix, quiet=True, internal=True)
        self.rowcount = len(qargs)  # type: ignore

        if not quiet and self._timing:
            self._execute_stdout(start)

        return self.rowcount

    def _validate_args(self, qargs, num_placeholders) -> bool:
        """Validate args for executemany() are in supported structure."""
        if not isinstance(qargs, (list, tuple)):
            raise ValueError(
                "The 'qargs' argument must be a list or tuple of parameter sets."
            )
        for i, param_set in enumerate(qargs):
            if not isinstance(param_set, (tuple, list)):
                raise ValueError(f"Parameter set at index {i} must be a tuple or list.")
            if len(param_set) != num_placeholders:
                raise ValueError(
                    f"Parameter set at index {i} has {len(param_set)} elements; "
                    f"expected {num_placeholders}."
                )
            # Check for nested lists or tuples
            if any(isinstance(item, (tuple, list)) for item in param_set):
                raise ValueError(
                    f"Parameter set at index {i} contains nested structures, which are not allowed: {param_set}"
                )

        return True

    def _query(self, stmt: str, generator: bool = False):
        """Execute query.
        : return : the row count in resultset.
        """
        conn = self._get_conn()
        self._executed = stmt
        self._clear_result()
        conn.query(stmt, generator=generator)
        self._do_get_result()
        return self.rowcount

    def fetchone(self):
        """Fetch one row data.
        : return : one row data.
        """
        try:
            self._check_executed()

            if self._rows is None or self.rownumber >= len(self._rows):
                return None
            result = self._rows[self.rownumber]
            self.rownumber += 1
        except TypeError:
            raise pyrdp.ProgrammingError("attempting to use unexecuted cursor")
        return result

    def fetchmany(self, size=None):
        """
        Fetch number of the result rows.
        : return : size or arraysize data rows.
        """
        self._check_executed()
        if self._rows is None:
            return ()
        end = self.rownumber + (size or self.arraysize)
        result = self._rows[self.rownumber : end]
        self.rownumber = min(end, len(self._rows))
        return result

    def fetchall(self):
        """
        Fetch all the rows.
        : return : all the result set.
        """
        self._check_executed()
        if self._rows is None:
            return ()

        if self.rownumber:
            result = self._rows[self.rownumber :]
        else:
            result = self._rows

        self.rownumber = len(result)
        return result

    def fetchall2(self):
        """
        Fetch all the rows.
        :return: A list containing all the rows in the result set.
        """
        self._check_executed()

        # If no rows are available, return an empty tuple
        if self._rows is None:
            return ()

        # If _rows is a generator, convert it to a list and store it
        if isinstance(self._rows, (list, tuple)):
            result = self._rows[self.rownumber :]
        else:
            # Consume the generator into a list
            result = list(self._rows)
            self._rows = result  # Cache the results for subsequent fetches

        # Update the current row pointer
        self.rownumber = len(result)

        return result

    @timeit
    def to_pandas(self, sql) -> pd.DataFrame:
        self.execute(sql, quiet=True)
        columns = [desc[0] for desc in self.description]
        return pd.DataFrame(self._rows, columns=columns)

    def scroll(self, value, mode="relative"):
        self._check_executed()
        if mode == "relative":
            r = self.rownumber + value
        elif mode == "absolute":
            r = value
        else:
            raise err.ProgrammingError("unknown scroll mode %s" % mode)

        if not (0 <= r < len(self._rows)):
            raise IndexError("out of range")
        self.rownumber = r

    def _get_execute_time(self, start: Optional[datetime]) -> float:
        if start:
            end = datetime.now()
            runtime = round((end - start).total_seconds(), 2)
        elif self.connection is not None:
            runtime = self.connection.current_execute_time
        else:
            runtime = -1  # executime unknown
        return runtime

    def _nextset(self, unbuffered=False):
        """Get the next query set."""
        conn = self._get_db()
        current_result = self._result
        if current_result is None or current_result is not conn._result:
            return None
        if not current_result.has_next:
            return None

        self._result = None
        self._clear_result()
        conn.next_result(unbuffered=unbuffered)
        if conn._result:
            self._do_get_result()
        return True

    def nextset(self):
        return self._nextset(unbuffered=False)

    def _do_get_result(self):
        """
        Get result after execute.
        """
        conn = self._get_conn()
        self._result = result = conn._result
        self.rowcount = result.affected_rows
        self.description = result.description
        self.err = result.err
        self._rows = result.rows
        if self.err:
            if (
                self.err.find("Got more than one row returned for the internal query")
                != -1
            ):
                pass
            else:
                err_msg = f"Exception: {self.err}, original query: {self._executed}"
                print(err_msg)
                raise pyrdp.RDBError(err_msg)

    def _trim(self, s):
        if s is None or s == "":
            return ""
        # 左侧空格
        while s[:1] == " ":
            s = s[1:]
        # 右侧空格
        while s[-1:] == " ":
            s = s[:-1]
        return s

    def _get_conn(self) -> "Connection":
        """Get the connection to execute query operation.

        Return : connection.
        """
        if not self.connection:
            raise pyrdp.ProgrammingError("Cursor closed")
        return self.connection

    def _check_executed(self):
        """Check if query is executed."""
        if not self._executed:
            raise pyrdp.ProgrammingError("execute() first")

    def _clear_result(self):
        """
        Clear last result info.
        """
        self.rownumber = 0
        self._result = None
        self.rowcount = 0
        self.description = None
        self.lastrowid = None
        self._rows = None

    def set_arraysize(self, size):
        """Set cursor arraysize.

        :param number size : arraysize of the cursor.
        """
        self.arraysize = size

    def get_arraysize(self):
        """Get cursor arraysize.

        Return : arraysize of the cursor.
        """
        return self.arraysize

    def callproc(self, procname, args=()):
        """Does nothing, required by DB API."""
        log.warning("callproc not implemented yet.")

    def setinputsizes(self, *args):
        """Does nothing, required by DB API."""
        log.warning("setinputsizes not implemented yet.")

    def setoutputsizes(self, *args):
        """Does nothing, required by DB API."""
        log.warning("setoutputsizes not implemented yet.")

    def _federation_exists(self, name: str, unbuffered: bool = False) -> bool:
        """Check if a federation name exists."""
        fed_name = name.upper()
        stmt = f"select * from FEDERATIONS where FEDERATION_NAME='{fed_name}';"
        self.execute(stmt, quiet=True)
        if not unbuffered:
            if self.rowcount == 1:
                return True
        else:
            self.fetchall()
            if self.rownumber == 1:
                return True
        return False

    def has_federation(self, name: str) -> bool:
        """Check if a federation name exists."""
        return self._federation_exists(name)

    def _connector_exists(self, name: str, unbuffered: bool = False) -> bool:
        """Check if a named connector exists."""
        connector_name = name.upper()
        stmt = "select connector_name from system.connectors where connector_name='{}';".format(
            connector_name
        )
        self.execute(stmt, quiet=True)
        if not unbuffered:
            if self.rowcount == 1:
                return True
        else:
            self.fetchall()
            if self.rownumber == 1:
                return True
        return False

    def has_connector(self, name: str) -> bool:
        """Check if a named connector exists."""
        return self._connector_exists(name)

    def _table_exists(
        self, table: str, schema: Optional[str] = None, unbuffered: bool = False
    ) -> bool:
        """Generic helper method to check if a named table exists."""
        tbl = table.upper()
        stmt = "select count(*) from tables where table_name='{}'".format(tbl)
        if schema:
            stmt += f" and schema_name = '{schema.upper()}'"
        self.execute(stmt, quiet=True)
        table_count = self.fetchone()[0]
        if table_count == 1:
            return True
        return False

    def has_table(self, table: str, schema: Optional[str] = None) -> bool:
        """Check if a named table exists."""
        return self._table_exists(table, schema)

    def get_table_names(self, schema=None, **kwargs) -> list[str]:
        """get all columns info of a specified schema"""
        schema = schema or "system"
        schema = "'" + schema.upper() + "'"
        self.execute(
            f"SELECT TABLE_NAME FROM RAPIDS.SYSTEM.TABLES WHERE SCHEMA_NAME={schema};"
        )

        table_names = list([row[0] for row in self.fetchall()])
        return table_names


class SSCursor(Cursor):
    """
    Unbuffered Cursor, mainly useful for queries that return a lot of data,
    or for connections to remote servers over a slow network.
    Instead of copying every row of data into a buffer, this will fetch
    rows as needed. The upside of this is the client uses much less memory,
    and rows are returned much faster when traveling over a slow network
    or if the result set is very big.
    There are limitations, though. The wireline protocol doesn't support
    returning the total number of rows, so the only way to tell how many rows
    there are is to iterate over every row returned. Also, it currently isn't
    possible to scroll backwards, as only the current row is held in memory.
    """

    def close(self):
        conn = self.connection
        if conn is None:
            return

        if self._result is not None and self._result is conn._result:
            pass

        try:
            while self.nextset():
                pass
        finally:
            self.connection = None

    __del__ = close

    def _query(self, stmt, generator: bool = False):
        conn = self._get_conn()
        self._executed = stmt
        self._clear_result()
        conn.query(stmt, unbuffered=True, generator=generator)
        self._do_get_result()
        return self.rowcount

    def nextset(self):
        return self._nextset(unbuffered=True)

    def read_next(self):
        """
        Read next row
        """
        return self._result.read_next_row_unbuffered()

    def fetchone(self) -> Optional[tuple]:
        """
        Fetch one row data.
        : return : one row data.
        """
        self._check_executed()
        row = self.read_next()
        if row is None:
            return None
        # result = self._rows[self.rownumber]
        self.rownumber += 1
        return row

    def fetchall(self):
        """
        Fetch all, as per MySQLdb. Pretty useless for large queries, as
        it is buffered. See fetchall_unbuffered(), if you want an unbuffered
        generator version of this method.
        """
        return list(self.fetchall_unbuffered())

    def fetchall_unbuffered(self):
        """
        Fetch all, implemented as a generator, which isn't to standard,
        however, it doesn't make sense to return everything in a list, as that
        would use ridiculous memory for large result sets.
        """
        return iter(self.fetchone, None)

    def scroll(self, value, mode="relative"):
        self._check_executed()

        if mode == "relative":
            if value < 0:
                raise err.NotSupportedError(
                    "Backwards scrolling not supported by this cursor"
                )

            for _ in range(value):
                self.read_next()
            self.rownumber += value
        elif mode == "absolute":
            if value < self.rownumber:
                raise err.NotSupportedError(
                    "Backwards scrolling not supported by this cursor"
                )

            end = value - self.rownumber
            for _ in range(end):
                self.read_next()
            self.rownumber = value
        else:
            raise err.ProgrammingError("unknown scroll mode %s" % mode)

    def fetchmany(self, size=None):
        """Fetch many"""
        self._check_executed()
        if size is None:
            size = self.arraysize

        rows = []
        for i in range_type(size):
            row = self.read_next()
            if row is None:
                break
            rows.append(row)
            self.rownumber += 1
        return rows

    def has_federation(self, name: str) -> bool:
        """Check if a federation name exists."""
        return self._federation_exists(name, unbuffered=True)

    def has_connector(self, name: str) -> bool:
        """Check if a named connector exists."""
        return self._connector_exists(name, unbuffered=True)

    def has_table(self, table: str, schema: Optional[str] = None) -> bool:
        """Check if a named table exists."""
        return self._table_exists(table, schema, unbuffered=True)


def convert_paramstyle(style, query, args):
    # I don't see any way to avoid scanning the query string char by char,
    # so we might as well take that careful approach and create a
    # state-based scanner.  We'll use int variables for the state.
    OUTSIDE = 0  # outside quoted string
    INSIDE_SQ = 1  # inside single-quote string '...'
    INSIDE_QI = 2  # inside quoted identifier   "..."
    INSIDE_ES = 3  # inside escaped single-quote string, E'...'
    INSIDE_PN = 4  # inside parameter name eg. :name
    INSIDE_CO = 5  # inside inline comment eg. --
    INSIDE_DQ = 6  # inside escaped dollar-quote string, $$...$$

    in_quote_escape = False
    in_param_escape = False
    placeholders = []
    output_query = []
    param_idx = map(lambda x: "$" + str(x), count(1))
    state = OUTSIDE
    prev_c = None
    for i, c in enumerate(query):
        next_c = query[i + 1] if i + 1 < len(query) else None

        if state == OUTSIDE:
            if c == "'":
                output_query.append(c)
                if prev_c == "E":
                    state = INSIDE_ES
                else:
                    state = INSIDE_SQ
            elif c == '"':
                output_query.append(c)
                state = INSIDE_QI
            elif c == "-":
                output_query.append(c)
                if prev_c == "-":
                    state = INSIDE_CO
            elif c == "$":
                output_query.append(c)
                if prev_c == "$":
                    state = INSIDE_DQ
            elif style == "qmark" and c == "?":
                output_query.append(next(param_idx))
            elif style == "numeric" and c == ":" and next_c not in ":=" and prev_c != ":":
                # Treat : as beginning of parameter name if and only
                # if it's the only : around
                # Needed to properly process type conversions
                # i.e. sum(x)::float
                output_query.append("$")
            elif style == "named" and c == ":" and next_c not in ":=" and prev_c != ":":
                # Same logic for : as in numeric parameters
                state = INSIDE_PN
                placeholders.append("")
            elif style == "pyformat" and c == "%" and next_c == "(":
                state = INSIDE_PN
                placeholders.append("")
            elif style in ("format", "pyformat") and c == "%":
                style = "format"
                if in_param_escape:
                    in_param_escape = False
                    output_query.append(c)
                else:
                    if next_c == "%":
                        in_param_escape = True
                    elif next_c == "s":
                        state = INSIDE_PN
                        output_query.append(next(param_idx))
                    else:
                        raise err.InterfaceError(
                            "Only %s and %% are supported in the query."
                        )
            else:
                output_query.append(c)

        elif state == INSIDE_SQ:
            if c == "'":
                if in_quote_escape:
                    in_quote_escape = False
                else:
                    if next_c == "'":
                        in_quote_escape = True
                    else:
                        state = OUTSIDE
            output_query.append(c)

        elif state == INSIDE_QI:
            if c == '"':
                state = OUTSIDE
            output_query.append(c)

        elif state == INSIDE_ES:
            if c == "'" and prev_c != "\\":
                # check for escaped single-quote
                state = OUTSIDE
            output_query.append(c)

        elif state == INSIDE_DQ:
            if c == "$" and prev_c == "$":
                state = OUTSIDE
            output_query.append(c)

        elif state == INSIDE_PN:
            if style == "named":
                placeholders[-1] += c
                if next_c is None or (not next_c.isalnum() and next_c != "_"):
                    state = OUTSIDE
                    try:
                        pidx = placeholders.index(placeholders[-1], 0, -1)
                        output_query.append("$" + str(pidx + 1))
                        del placeholders[-1]
                    except ValueError:
                        output_query.append("$" + str(len(placeholders)))
            elif style == "pyformat":
                if prev_c == ")" and c == "s":
                    state = OUTSIDE
                    try:
                        pidx = placeholders.index(placeholders[-1], 0, -1)
                        output_query.append("$" + str(pidx + 1))
                        del placeholders[-1]
                    except ValueError:
                        output_query.append("$" + str(len(placeholders)))
                elif c in "()":
                    pass
                else:
                    placeholders[-1] += c
            elif style == "format":
                state = OUTSIDE

        elif state == INSIDE_CO:
            output_query.append(c)
            if c == "\n":
                state = OUTSIDE

        prev_c = c

    if style in ("numeric", "qmark", "format"):
        vals = args
    else:
        vals = tuple(args[p] for p in placeholders)

    return "".join(output_query), vals
